from .dimension_controls import DimensionControls
from .histogram import HistogramFigure
from .image import ImageFigure
from .palette import PaletteControls
from .statistics import StatisticsFigure
from .value_controls import ValueRadios
from .profile import ProfileFigure

__all__ = ["DimensionControls", "HistogramFigure", "ImageFigure", "PaletteControls", "StatisticsFigure", "ValueRadios", "ProfileFigure"]


from typing import List, Union, Optional, Dict, Tuple

import ismrmrd
from bokeh.models import Div
from ismrmrd.file import Container
from pyscivis.visualizer.extensions import NestedExtension
from pyscivis.visualizer.plots import ImagePlot, ComplexPlot
from pyscivis.visualizer.plots.table_plot import TablePlot

from .image_parser import ImageParser
from .acquisition_parser import AcquisitionParser
from .header_parser import HeaderParser

from .tree_util import extract_data_leaves, extract_tree_structure


class MRDExtension(NestedExtension):
    """
    Extension to handle ISMRMRD-type h5-files.
    """

    alias = "mrd"
    supported_files = ".h5"
    file_description = "ISMRMRD-files"
    multiple = True

    def create_plot(self,
                    leaf_name: str,
                    container_name: str,
                    container_type: str,
                    loading_div: Optional[Div] = None
                    ) -> Union[ImagePlot, ComplexPlot, TablePlot]:
        """
        Create plot by accessing the ismrmrd.Container directly, based on name and type of container.

        Args:
            leaf_name: Full ID of clicked leaf, not used  because we address files by their parent container and type.
            container_name: Name of the ismrmrd container.
            container_type: Type of the ismrmrd container.
            loading_div: If a loading div is to be used for progress notification.

        Returns:
            An ImagePlot or ComplexPlot.
        """
        try:
            container = self.data_leaves[container_name]  # retrieve h5-container
        except KeyError:
            raise ValueError(f"File-Leaf '{container_name}' does not occur in dict of File-Leaves ")

        parser_conf = self.config.parser
        plot_config = self.config.image
        # parse container depending on its header and create a plot
        if container_type == "images":
            data = container.images
            header = container.images[0].getHead()
            parsed_img = ImageParser.parse(data, header, parser_conf, loading_indicator=loading_div)
            return ImagePlot(parsed_img, plot_config, loading_indicator=loading_div)
        elif container_type == "acquisitions":
            data = container.acquisitions
            header = container.header
            parsed_acq = AcquisitionParser.parse(data, header, parser_conf, loading_indicator=loading_div)
            return ComplexPlot(parsed_acq, plot_config, loading_indicator=loading_div)
        elif container_type == "header":
            header_tree = HeaderParser.parse(container.header, loading_indicator=loading_div)
            return TablePlot(header_tree, loading_indicator=loading_div)
        elif container_type == "waveforms":
            # examples of ismrmrd waveforms are hard to come by
            # will be implemented once actually needed by anyone
            raise NotImplementedError
        else:  # unknown, don't do anything
            pass

    def get_valid_leaf_type_combinations(self) -> List[str]:
        """
        Check what kind of data a container contains and combine this type with the name of the parent container.

        Returns:
            A list of valid combinations that can be used to identify a specific data-container.

        """
        #  create valid leaf-name + type combinations
        valid_combinations = list()
        for name, container in self.data_leaves.items():
            valid_types: List[str] = list()

            if container.has_images():
                valid_types.append("images")
            if container.has_acquisitions():
                valid_types.append("acquisitions")
            if container.has_header():
                valid_types.append("header")
            if container.has_waveforms():
                valid_types.append("waveforms")
            valid_combinations.extend([name + ";" + _type for _type in valid_types])
        return valid_combinations

    @staticmethod
    def file_handler(path: str) -> ismrmrd.File:
        return ismrmrd.File(path, mode="r")

    @staticmethod
    def get_tree_structure(path: Tuple[str]) -> List[Dict[str, str]]:
        return extract_tree_structure(path, MRDExtension.file_handler)

    @staticmethod
    def get_data_leaves(path: Tuple[str]) -> Dict[str, Container]:
        return extract_data_leaves(path, MRDExtension.file_handler)



from typing import Optional, Union

import numpy as np
from PIL import Image
from pyscivis.visualizer.dataclasses.parser import ParsedData

from pyscivis.visualizer.extensions import FlatExtension
from pyscivis.visualizer.plots import ImagePlot, ComplexPlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv


class PILExtensionRGBA(FlatExtension):

    alias = "pil_rgba"
    supported_files = (".jpg", ".jpeg", ".png", ".gif", ".bmp")
    file_description = "PIL-supported RGBA-Image"
    multiple = False

    def create_plot(self, loading_div: Optional[LoadingDiv] = None) -> Union[ImagePlot, ComplexPlot]:
        img: Image.Image
        with Image.open(self.path) as img:
            img = img.convert('RGBA')
            np_arr = np.asarray(img).view(dtype=np.uint32).squeeze()
        data = ParsedData(
            data=np_arr,
            dim_names=["y", "x"],
            dim_lengths=np_arr.shape,
            dim_units=["px", "px"]
        )
        self.config.image.ds_min_elements = 1000
        return ImagePlot(data, self.config.image, use_rgba=True)


from time import perf_counter
from functools import partial

from bokeh.models.widgets import Slider

# Most of the credit goes to Matt_Agee (https://discourse.bokeh.org/t/server-side-throttled-slider/3212)

class ThrottledSlider(Slider):
    """
    Extends the bokeh Slider to provide throttling of callbacks provided to on_update
    """

    #__implementation__ = "throttled_slider.ts"

    def __init__(self, base_delay, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._cb_set = False
        self._base_delay = base_delay
        self._delay = base_delay

    def on_change(self, attr, callback):
        """
        Register a server side callback to be executed when slider attributes change

        Args:
            attr: Attribute that changed.
            callback: Callback to fire.
        """
        update_function = self._callback_wrapper(callback)
        super().on_change(attr, partial(self._callback_setter, update_function))

    def _callback_setter(self, update_function, attr, old, new):
        if not self._cb_set:
            self._document.add_timeout_callback(partial(update_function, attr, old), self._delay)
            self._cb_set = True

    def _callback_wrapper(self, callback):
        """
        Wraps callback to use self.value instead of new.

        This ensures the callback has the most recent slider value at execution.
        Also modifies the delay between callback execution to account for time taken
        performing the callback's work.
        """

        def throttled_cb(*args):
            start_time = perf_counter()
            attr, old = args
            callback(attr, old, self.value)

            # Modify delay to account for time taken executing callback
            self._delay = max(self._base_delay - 1000 * (perf_counter() - start_time), 50)
            self._cb_set = False

        return throttled_cb



from typing import List, Dict, Optional
import uuid
from xml.etree.ElementTree import Element

from defusedxml import ElementTree
from ismrmrd.xsd.ismrmrdschema.ismrmrd import ismrmrdHeader as AcquisitionsHeader
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv


class HeaderParser:
    @staticmethod
    def parse(header: AcquisitionsHeader,
              loading_indicator: Optional[LoadingDiv] = None
              ) -> Dict[str, List]:
        xml_tree = ElementTree.fromstring(header.toXML())
        column_dict = dict(name=list(), value=list())
        HeaderParser.create_list(column_dict, xml_tree)
        return column_dict

    @staticmethod
    def create_list(column_dict, node, depth=0):
        elem: Element
        for elem in node:
            name = elem.tag[30:]
            name = f"{depth*'——'} {name}"
            column_dict['name'].append(name)
            if not len(elem):
                column_dict['value'].append(elem.text)
            else:
                column_dict['value'].append("")
                HeaderParser.create_list(column_dict, elem, depth+1)



from tornado import gen

from bokeh.models import Div
from bokeh.layouts import row


class LoadingDiv:
    """
    Wraps a loading screen and a text-Div to display a status message.
    """

    def __init__(self, text: str = "Placeholder") -> None:
        """
        Set up the loading-Div and status-Div.

        Args:
            text: Initial text to be displayed in the status message.
        """
        loading_screen = Div(text="""<div class="loading style-2"><div class="loading-wheel"></div></div>""")
        self.text_div = Div(text=text, css_classes=["loading-text"])
        self.layout = row(loading_screen, self.text_div, visible=False)  # hidden by default

    def set_text(self, text: str) -> None:
        """
        Set the status text from any thread (main or other).

        Args:
            text: Text to display.

        Note:
            The text will not be set if the main-thread is not passing control to bokeh/is busy.
        """
        @gen.coroutine
        def _set_text():
            self.layout.visible = True
            self.text_div.text = text
        self.text_div.document.add_next_tick_callback(_set_text)

    def show(self) -> None:
        @gen.coroutine
        def _show():
            self.layout.visible = True
        self.text_div.document.add_next_tick_callback(_show)

    def hide(self) -> None:
        @gen.coroutine
        def _hide():
            self.layout.visible = False
        self.text_div.document.add_next_tick_callback(_hide)



import glob
import os

from PIL import Image
import numpy as np


def get_images(width, height):
    """
    Collect all png-images in the same folder as this script, resize them and convert them to numpy-arrays.

    Args:
        width: Width for images.
        height: Height for images.

    Returns:
        3-dimensional numpy-array with shape (image-count, height, width).
    """
    image_paths = glob.glob(os.path.join(os.path.dirname(__file__), "*.png"))
    image_series = []
    img: Image.Image
    for i, p in enumerate(image_paths):
        with Image.open(p) as img:
            img = img.resize((width, height)).convert('L')
            image_series.append(np.asarray(img))

    return np.array(image_series)


if __name__ == "__main__":
    print(get_images(100, 200).shape)



import logging
import os
from typing import Optional, Tuple, Union, List, Sequence

from pyscivis.visualizer.dataclasses.config import Config
from pyscivis.visualizer.extensions import FlatExtension, NestedExtension

#########################################################################
#                      !!!Enable Extensions here!!!                     #
#                                                                       #
# Import your Extension-Class here and put it into 'enabled_extensions' #
#########################################################################

from .mrd import MRDExtension
from .pil.pil_flat import PILExtensionFlat
from .pil.pil_nested import PILExtensionNested
from .pil.pil_rgba import PILExtensionRGBA

enabled_extensions = [MRDExtension, PILExtensionFlat, PILExtensionNested, PILExtensionRGBA]

#########################################################################
#                                                                       #
#########################################################################

"""
This module includes functions used to validate and use
custom extensions.
"""

logger = logging.getLogger(__name__)


def get_extension_handler(path: Union[str, Sequence[str]],
                          extension_name: Optional[str],
                          config: Config
                          ) -> Union[FlatExtension, NestedExtension]:
    """
    Get the ExtensionHandler for a file.

    Use the extension name to choose an extension, if not specified use the filename-extension, e.g., ".jpg".

    Args:
        path: Path(s) of file(s) to be handled.
        extension_name: Name of extension to be used to handle the file.
        config: Config object.

    Returns:
        An ExtensionHandler (subclassing either FlatExtension or NestedExtension).
    """
    if extension_name is None:
        return get_extension_handler_from_path(path, config)
    else:
        return get_extensions_handler_from_name(extension_name, path, config)


def get_extension_handler_from_path(path: Union[str, Sequence[str]],
                                    config: Config
                                    ) -> Union[FlatExtension, NestedExtension]:
    """
    Get the appropriate ExtensionHandler for the specified file(-path).

    Args:
        path: Path(s) of file(s) to be handled.
        config: Config object.

    Returns:
        An ExtensionHandler (subclassing either FlatExtension or NestedExtension).

    Raises:
        ValueError: If the specified file (first file for Sequences) does not have a file-extension.
        NotImplementedError: If no extensions supports the path's file-extension.
        IndexError: If the file-extension is not unique to an extension, but multiple extensions support it.
    """

    file_ext_name: str
    # use first path for detecting the extension if path is a Sequence of strings
    if not isinstance(path, str):
        _, file_ext_name = os.path.splitext(path[0])
    else:
        _, file_ext_name = os.path.splitext(path)

    if not file_ext_name:
        raise ValueError("Could not find file-extension in path.")

    file_ext_name_no_dot: Optional[str] = None
    if file_ext_name.startswith("."):
        file_ext_name_no_dot = file_ext_name[1:]

    valid_extensions = list()
    for ext in enabled_extensions:
        if file_ext_name in ext.supported_files or file_ext_name_no_dot in ext.supported_files:
            valid_extensions.append(ext)

    if len(valid_extensions) == 1:
        return valid_extensions[0](path, config)
    elif len(valid_extensions) == 0:
        raise NotImplementedError(f"'{file_ext_name}'-files are not currently supported.")
    else:  # len(valid_extensions) > 1
        raise IndexError(f"'{file_ext_name}'-files are supported by multiple extensions.")


def get_extensions_handler_from_name(extension_name: str,
                                     path: Union[str, Sequence[str]],
                                     config: Config
                                     ) -> Union[FlatExtension, NestedExtension]:
    """
    Get the appropriate ExtensionHandler for the specified file(path) using the specified extension.

    Args:
        extension_name: Name of extensions to be used.
        path: Path(s) of file(s) to be handled.
        config: Config object.

    Returns:
        An ExtensionHandler (subclassing either FlatExtension or NestedExtension).

    Raises:
        NotImplementedError: If no extension with the specified name exists.
    """
    for ext in enabled_extensions:
        if ext.alias == extension_name:
            return ext(path, config)
    raise NotImplementedError(f"Extension with name '{extension_name}' does not exist.")


def extension_allows_multiple(extension_name: str) -> bool:
    """
    Check if the specified extension can handle multiple files or not.

    Args:
        extension_name: Name of extension to be checked.

    Returns:
        True if extension can handle multiple files, else False.

    Raises:
        ValueError: If no extension with specified name exists.
    """
    for ext in enabled_extensions:
        if ext.alias == extension_name:
            return ext.multiple
    raise ValueError(f"Extension with name '{extension_name}' does not exist.")


def get_supported_files() -> Tuple[Tuple[str, str, str], ...]:
    """
    Compile the supported file-extensions of all enabled extensions.

    Returns:
        A tuple of (extension-alias, extension-description, extension-filetypes)-tuples.

    Raises:
        TypeError: If an enabled extension did not set 'supported_files' correctly.
    """
    if not enabled_extensions:
        return ("NO-EXTENSION-LOADED", "NO-EXTENSION-LOADED", ".NO-EXTENSION-LOADED"),

    supported_files = list()
    for ext in enabled_extensions:
        if isinstance(ext.supported_files, str):
            file_ext = ext.supported_files if ext.supported_files.startswith((".", "*")) else f".{ext.supported_files}"
            supported_files.append((ext.alias, ext.file_description, file_ext))
        elif isinstance(ext.supported_files, Sequence):
            file_ext_list = [f if f.startswith((".", "*")) else f".{f}" for f in ext.supported_files]
            supported_files.append((ext.alias, ext.file_description, " ".join(file_ext_list)))
        else:
            raise TypeError(f"Extension {ext.__package__} has invalid type for variable 'supported_files. "
                            f"Expected type Union[str, Sequence], got type {type(ext.supported_files)}")

    return tuple(supported_files)


def get_extension_names_and_descriptions() -> Tuple[Tuple[str, str], ...]:
    """
    Get the names of all extensions.

    Returns: (extension-name, extension-description)-Tuple.
    """
    ret: List[Tuple[str, str]] = list()
    for ext in enabled_extensions:
        ret.append((ext.alias, ext.file_description))
    return tuple(ret)


def check_extension_validity() -> None:
    """
    Check that all enabled extensions fulfil basic requirements for extensions to function correctly.

    I.e., it is checked whether the fields 'file_description' and 'supported_files' exist and that the
    ExtensionHandler is a subclass of FlatExtension or NestedExtension.
    All invalid extensions are disabled.

    Returns:
        True if all extensions are valid, else false.
    """
    logger.info(f"Checking enabled extensions {[ext.__name__ for ext in enabled_extensions]} for validity.")

    if not enabled_extensions:
        logger.warning("No extension was enabled. The application will be missing most functionality.")

    errors = dict()
    for ext in enabled_extensions:
        errors[ext] = []

        if isinstance(ext.alias, property):
            errors[ext].append("missing required field 'alias'")
        elif not isinstance(ext.alias, str):
            errors[ext].append("field 'alias' is not string")

        if isinstance(ext.supported_files, property):
            errors[ext].append("missing required field 'supported_files'")
        elif not isinstance(ext.supported_files, (str, tuple)):
            errors[ext].append("field 'supported_files' is not string or tuple")

        if isinstance(ext.file_description, property):
            errors[ext].append("missing required field 'file_description'")
        elif not isinstance(ext.file_description, str):
            errors[ext].append("field 'file_description' is not string")

        if isinstance(ext.multiple, property):
            errors[ext].append("missing required field 'multiple'")
        elif not isinstance(ext.multiple, bool):
            errors[ext].append("field 'multiple' is not boolean")

        if not issubclass(ext, (FlatExtension, NestedExtension)):
            errors[ext].append("Extension does not subclass 'FlatExtension' or 'NestedExtension'")

    if any(errors.values()):
        logger.warning("Not all enabled extensions are configured correctly:")
        for ext, msgs in errors.items():
            if not msgs:
                continue

            logger.warning(f"Extension {ext.__name__} has the following errors and will be disabled: {', '.join(msgs)}.")
            enabled_extensions.remove(ext)
    else:
        logger.info("All Extensions were successfully validated.")



from warnings import warn

import numpy as np
from ismrmrd import ImageHeader
from ismrmrd.file import Images

from pyscivis.visualizer.dataclasses.config import ParserConfig
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv

from .base_parser import BaseParser
from pyscivis.visualizer.dataclasses.parser import ParsedData

default_image_dimensions = ("Instances", "Channels", "Slices", "Y", "X")  # Slices ^= Z


class ImageParser(BaseParser):
    """
    A parser to extract images and their metadata from either ismrmrd-Images or numpy-arrays.
    """

    @staticmethod
    def parse_with_header(images: Images,
                          header: ImageHeader,
                          config: ParserConfig,
                          printer: LoadingDiv
                          ) -> ParsedData:
        """
        Take Images-data from ismrmrd-files and compile images.

        First, the image-data is retrieved from the Images-object and converted to a np.ndarray.
        Then, optional FoV-data is extracted from the header. This includes dimension lengths and units
        which usually only are set for kX, kY and kZ.

        Args:
            images: An ismrmrd-Images object.
            header: An ismrmrd-ImageHeader containing relevant metadata.
            config: An object containing the parser configuration.
            printer: A callable that accepts a string and displays it.

        Returns:
            A ParsedData object.
        """
        images_data = np.array(images.data)
        if images_data.ndim != 5:
            warn(f"ISMRMRD-images_data has unsupported number of dimensions: {images_data.ndim}; \
                please create a Github-Issue.", UserWarning)

        # the existence of a header implies a standard ismrmrd image -> use default_dimensions
        names = default_image_dimensions
        try:
            _fov = tuple(header.field_of_view)[::-1]
        except AttributeError:
            _fov = [0]

        if _fov.count(0):  # if FoV contains any 0s/was nulled/does not exist
            lengths = images_data.shape
            units = tuple(["pixel"]*images_data.ndim)
        else:
            lengths = images_data.shape[:-len(_fov)] + _fov
            _amount_fillers = images_data.ndim - len(_fov)
            units = tuple(["pixel"]*_amount_fillers) + tuple(["mm"]*len(_fov))
        return ParsedData(images_data, names, lengths, units)




import os

from bokeh.io import curdoc

from pyscivis.visualizer.dataclasses.config import load_config
from pyscivis.visualizer.standalone.standalone import MainComponent
from pyscivis.arguments import parse_args

config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "config.toml")  # TODO: make this nicer
config = load_config(config_path)
args = parse_args()

doc = curdoc()
doc.title = "pyscivis"
doc.add_root(MainComponent(config, **vars(args)).layout)



from functools import wraps, partial
from typing import Union, Dict, Optional

from pyscivis.visualizer.dataclasses.state import State
from pyscivis.visualizer.plots import ImagePlot, ComplexPlot
from pyscivis.visualizer.plots.components.value_controls import LABEL_FUNC_MAP as VALID_VALUE_KINDS


def on_next_tick(func):
    """ Decorator to allow doc-modification outside the main thread. """

    @wraps(func)
    def _impl(self, *args, **kwargs):
        return self.plot.document.add_next_tick_callback(partial(func, self, *args, **kwargs))

    return _impl


class PlotWrapper:
    """ This class is used to keep a reference to displayed plots and modify them programmatically. """

    def __init__(self, plot: Union[ImagePlot, ComplexPlot]) -> None:
        """
        Save plot and the plot's bokeh-document for future use.

        Args:
            plot: Plot object to be wrapped.
        """
        self.plot = plot

    @on_next_tick
    def set_state(self, state: State) -> None:    # pragma: nocover
        """
        Set the state of the plot manually.

        Args:
            state: A State object.
        """
        self._set_state(state)

    def _set_state(self, state: State) -> None:
        self.plot.set_state(state)

    def get_state(self) -> State:
        """
        Get the state of the plot manually.

        Returns:
            A State object.
        """
        return self.plot.get_state()

    @on_next_tick
    def set_dimension(self,
                      name: str,
                      value: int
                      ) -> None:    # pragma: nocover
        """
        Change a dimension if the dimension-name is valid.

        Automatically caps the values into range of (start, end) of slider.

        Args:
            name: Name of the dimension.
            value: New dimension value.
        """
        self._set_dimension(name, value)

    def _set_dimension(self,
                       name: str,
                       value: int
                       ) -> None:
        if self._is_valid_dimension_name(name):
            for dim in self.plot.dim_controls.dim_sliders:
                if dim.name == name:
                    min_, max_ = dim.slider.start, dim.slider.end
                    if value < min_:
                        value = min_
                    elif value > max_:
                        value = max_
                    dim.slider.value = value

    @on_next_tick
    def set_x_axis(self, name: str) -> None:    # pragma: nocover
        """
        Change the dimension displayed by the x-axis if the dimension-name is valid.

        Args:
            name: Name of the dimension.
        """
        self._set_x_axis(name)

    def _set_x_axis(self, name: str) -> None:
        """Set the x-axis value to name"""
        if self._is_valid_dimension_name(name):
            self.plot.dim_controls.axis_select.x_select.value = name

    @on_next_tick
    def set_y_axis(self, name: str) -> None:    # pragma: nocover
        """
        Change the dimension displayed by the y-axis if the dimension-name is valid.

        Args:
            name: Name of the dimension.
        """
        self._set_y_axis(name)

    def _set_y_axis(self, name: str) -> None:
        """Set the x-axis value to name"""
        if self._is_valid_dimension_name(name):
            self.plot.dim_controls.axis_select.y_select.value = name

    def _is_valid_dimension_name(self, name: str) -> bool:
        """
        Check if name is a valid dimension identifier and prints valid names if False.

        Args:
            name: Name of the dimension.

        Returns:
            True if name is valid, else False.
        """
        axis_select = self.plot.dim_controls.axis_select
        if name in axis_select.options:
            return True

        print("Specified dimension does not exist, possible dimensions are: " +
              ", ".join(list(axis_select.options.keys())))
        return False

    @on_next_tick
    def set_value_kind(self, name: str) -> None:    # pragma: nocover
        """
        Set the value kind of a ComplexPlot.

        Args:
            name: Name of the new value kind ("Real", "Imaginary", "Abs", "Phase")
        """
        self._set_value_kind(name)

    def _set_value_kind(self, name: str) -> None:
        if not isinstance(self.plot, ComplexPlot):
            print("Plot does not have complex values.")
            return
        valid = list(VALID_VALUE_KINDS)
        if name not in valid:
            print(f"Value kind {name} has to be one of {valid}.")
            return

        self.plot.value_controls.set_active_handler(name)

    def get_stats(self) -> Dict:
        """
        Get a copy of the current image-statistics.

        Returns:
            A dictionary containing statistics.
        """
        return self.plot.statistics.stats.copy()

    @on_next_tick
    def toggle_controls(self,
                        left: Optional[bool] = None,
                        bottom: Optional[bool] = None,
                        right: Optional[bool] = None
                        ) -> None:  # pragma: nocover
        """
        Toggle the visibility of the left, bottom or right control-columns.

        Args:
            left: Whether to hide the left column.
            bottom: Whether to hide the bottom column.
            right: Whether to hide the right column.

        Notes:
            Broken because of bokeh as of bokeh-2.3.2: https://github.com/bokeh/bokeh/issues/11339
            TODO: Updating bokeh once 2.4 releases might fix this
        """
        self._toggle_controls(left, bottom, right)

    def _toggle_controls(self,
                         left: Optional[bool] = None,
                         bottom: Optional[bool] = None,
                         right: Optional[bool] = None
                         ) -> None:
        doc = self.plot.document
        if left is not None:
            doc.select_one({"name": "col_left"}).visible = left
        if bottom is not None:
            doc.select_one({"name": "col_bottom"}).visible = bottom
        if right is not None:
            doc.select_one({"name": "col_right"}).visible = right

    @on_next_tick
    def toggle_fit_to_frame(self, active: Optional[bool] = None) -> None:    # pragma: nocover
        """
        Toggle the fit-to-frame-toggle.

        If 'active' is unspecified, the toggle's state will be flipped, else 'active' is applied.

        Args:
            active: Boolean for activating/disabling the fit_to_frame_toggle.
        """
        self._toggle_fit_to_frame(active)

    def _toggle_fit_to_frame(self, active: Optional[bool] = None) -> None:
        if active is not None:
            self.plot.fit_to_frame_toggle.active = active
        else:
            self.plot.fit_to_frame_toggle.active = not self.plot.fit_to_frame_toggle.active



from functools import partial
import threading
from typing import Optional

import numpy as np
from bokeh.io import curdoc
from bokeh.models import ColumnDataSource, glyphs, Label, WheelZoomTool
from bokeh.plotting import figure

delayed_cb: Optional[threading.Timer] = None


class HistogramFigure:
    """
    Manages the creation of a histogram-figure and its bounds.
    """
    def __init__(self,
                 data: np.ndarray,
                 min_: float,
                 max_: float,
                 bins: int = 20,
                 max_calc_size: int = 4000*4000,  # 16kk ~= 0.5s calculation @ i7-3770,
                 enable_dynamic_rebinning: bool = True
                 ) -> None:
        """
        Set up the figure and its initial two histogram-glyphs with placeholder data.

        Args:
            data: N-dimensional numpy-array containing data to be binned into a histogram.
            min_: Minimum of the data, used to calculate x-range start.
            max_: Maximum of the data, used to calculate x-range end.
            bins: Amount of histogram bins.
            max_calc_size: Threshold of when not to calculate a histogram anymore.
            enable_dynamic_rebinning: Whether to automatically rebin the histogram if zoomed in/out or panned.
        """

        # we need min_max as parameters because the Histogram only has access to 1 frame which
        # is not enough for min_max because we want min_max to be constant for an entire image-series

        self.min_max = (0., 1.)
        self.max_calc_size = max_calc_size
        self.bins = bins

        tooltip = [
            ("amount", "@top")
        ]

        fig = figure(tools=["hover,reset,pan,box_zoom"], toolbar_location="left", tooltips=tooltip,
                     plot_height=200, sizing_mode="stretch_width", x_range=(0, 1),
                     y_range=(1e-1, 2*data.size), y_axis_type="log", y_axis_location="right",
                     min_border=10, css_classes=["histogram-figure"])
        wztool = WheelZoomTool(maintain_focus=False, dimensions="width")
        fig.add_tools(wztool)

        fig.toolbar.active_scroll = wztool
        fig.toolbar.active_drag = "auto"
        fig.toolbar.logo = None

        fig.y_range.bounds = (1e-1, 2*data.size)
        self.h_main: glyphs.Quad = fig.quad(bottom=1, left=[1], right=[1], top=[1],
                                            color="white", line_color="#3A5785", visible=False)
        self.h_selected: glyphs.Quad = fig.quad(bottom=1, left=[1], right=[1], top=[1],
                                                alpha=0.5, color="#3A5785", line_color=None, visible=False)

        self.no_display = Label(x=fig.width/2, y=fig.height/2, x_units="screen", y_units="screen",
                                text=f"Selection size is bigger than threshold of {max_calc_size}",
                                render_mode="css", border_line_color="darkgrey", border_line_alpha=1.0,
                                background_fill_color="white", background_fill_alpha=1.0,
                                text_align="center")

        fig.add_layout(self.no_display, place="center")

        fig.hover.renderers = [self.h_selected]
        fig.hover.mode = "vline"

        fig.xgrid.grid_line_color = None
        fig.yaxis.major_label_orientation = np.pi / 6
        fig.background_fill_color = "#fafafa"

        self.cur_data = [data, data]  # all, selection
        self.cur_bins = bins

        self.fig = fig
        self.layout = fig

        self.update_min_max(min_, max_)
        self.update(data, data)
        if enable_dynamic_rebinning:
            self._attach_range_events()

    def update(self,
               entire_img: np.ndarray,
               selection: np.ndarray
               ) -> None:
        """
        Update both histogram-glyphs with supplied data.

        Args:
            entire_img: N-dimensional numpy-array for the main-histogram-glyph.
            selection: N-dimensional numpy-array for the selected-histogram-glyph.
        """
        self.cur_data = [entire_img, selection]

        if entire_img.size*2 != self.fig.y_range.end:
            self.fig.y_range.end = entire_img.size*2
            self.fig.y_range.bounds = (1e-1, entire_img.size*2)

        if selection.size > self.max_calc_size:
            self.h_main.visible = False
            self.h_selected.visible = False
            self.no_display.visible = True
            return

        self.h_selected.visible = True
        self.no_display.visible = False

        sel_ds: ColumnDataSource = self.h_selected.data_source.data
        main_ds: ColumnDataSource = self.h_main.data_source.data

        x_start, x_end = self.fig.x_range.start, self.fig.x_range.end

        if entire_img.size > self.max_calc_size:
            self.h_main.visible = False
            selected_vals = selection.ravel()

            hist_selected, hedges = self.filtered_histogram(x_start, x_end, selected_vals, range=self.min_max, bins=self.cur_bins)

            # Need this here to avoid a delay in updates between all & selected
            sel_ds.update(left=hedges[:-1], right=hedges[1:], top=hist_selected)

        else:
            self.h_main.visible = True

            all_vals = entire_img.ravel()
            hist_all, hedges = self.filtered_histogram(x_start, x_end, all_vals, range=self.min_max, bins=self.cur_bins)

            if entire_img.shape == selection.shape:
                sel_ds.update(left=hedges[:-1], right=hedges[1:], top=hist_all)
                main_ds.update(left=hedges[:-1], right=hedges[1:], top=hist_all)
            else:
                selected_vals = np.array(selection).ravel()

                hist_selected, _ = self.filtered_histogram(x_start, x_end, selected_vals, range=self.min_max, bins=self.cur_bins)
                sel_ds.update(left=hedges[:-1], right=hedges[1:], top=hist_selected)
                main_ds.update(left=hedges[:-1], right=hedges[1:], top=hist_all)
                # Need this here to avoid a delay in updates between all & selected

    def update_selection(self, selection: np.ndarray) -> None:
        """
        Update only the selected-histogram-glyph.

        Args:
            selection:  N-dimensional numpy-array for the selected-histogram-glyph.
        """
        self.cur_data[1] = selection

        if selection.size > self.max_calc_size:
            self.h_selected.visible = False
            self.no_display.visible = True
            return

        vals = selection.ravel()
        hhist, hedges = np.histogram(vals, range=self.min_max, bins=self.cur_bins)
        sel_cds: ColumnDataSource = self.h_selected.data_source.data
        sel_cds.update(left=hedges[:-1], right=hedges[1:], top=hhist)
        self.h_selected.visible = True
        self.no_display.visible = False

    def update_min_max(self,
                       min_: float,
                       max_: float
                       ) -> None:
        """
        Update the bounds of the histogram plot and the range for future histogram calculations.

        Args:
            min_: Start value.
            max_: End value.
        """

        if np.isnan(min_) or np.isnan(max_) or min_ == 0 and max_ == 0:
            min_ = 0
            max_ = 1

        self.min_max = (min_, max_)
        pad = 0.1 * (max_ - min_)
        self.layout.x_range.update(start=min_-pad,
                                   end=max_+pad,
                                   bounds=(min_-pad, max_+pad),
                                   reset_start=min_-pad,
                                   reset_end=max_+pad)

    @staticmethod
    def filtered_histogram(start, end, *args, **kwargs):
        """
        Filters the edges and histogram-values to be within the specified start-end range.

        Useful for limiting the amount of drawn glyphs to the visible area.

        Args:
            start: Value before which the histogram is cut off.
            end: Value after which the histogram is cut off.
            *args: Arguments passed on to `np.histogram`.
            **kwargs: Keyword-arguments passed on to `np.histogram`.

        Returns:
            Tuple containing histogram heights and the edges (exactly like `np.histogram`).
        """
        hist, edges = np.histogram(*args, **kwargs)

        left_edges = edges[:-1]
        right_edges = edges[1:]
        # this does include the leftmost partial rect but not
        # the rightmost one
        mask = (start <= right_edges) & (left_edges < end)
        # so we find the rightmost edge here and set it to visible
        idx_shown = np.argmax(mask)  # find idx of first shown edge
        true_onwards = mask[idx_shown:]
        if all(true_onwards):
            mask = np.insert(mask, -1, True)  # make it visible
        else:
            idx_first_false = np.argmin(true_onwards)  # find idx of first non-visible edge
            mask = np.insert(mask, idx_first_false, True)  # make it visible

        #print(idx_first_false)

        valid_edges = edges[mask]
        valid_hist = hist[mask[:-1]]

        # if the last element gets filtered we need to remove the last histogram value
        if not mask[-1]:
            valid_hist = valid_hist[:-1]

        return valid_hist, valid_edges

    def _attach_range_events(self):
        """
        Attach a throttled callback listening to changes in the figure's x-range and rebinning if necessary.
        """
        # cur_zoom_level = self.min_max[1] - self.min_max[0]

        self.fig.x_range.on_change("start", lambda attr, old, new: self._range_change_throttled())
        self.fig.x_range.on_change("end", lambda attr, old, new: self._range_change_throttled())

    def _range_change_throttled(self):
        timeout = 0.01  # 10ms
        global delayed_cb
        if delayed_cb is not None:
            delayed_cb.cancel()
        threaded_rebin = partial(curdoc().add_next_tick_callback, self.rebin_histogram)
        delayed_cb = threading.Timer(timeout, threaded_rebin)
        delayed_cb.start()

    def rebin_histogram(self):
        # TODO: Possible improvement: Differentiate between zoom and pan, for pan don't recalc the histogram
        # nonlocal cur_zoom_level
        # self.min_max = start, end
        start, end = self.fig.x_range.start, self.fig.x_range.end
        new_x_width = end - start
        # if new_x_width == cur_zoom_level: #math.isclose(cur_range, cur_zoom_level, abs_tol=0):
        #     print("%s and %s are close" % (new_x_width, cur_zoom_level))
        #     return
        # print("%s and %s are not close" % (new_x_width, cur_zoom_level))
        orig_x_width = self.min_max[1] - self.min_max[0]
        self.cur_bins = round(orig_x_width / new_x_width * self.bins)
        self.update(*self.cur_data)



from typing import TypedDict, Optional, Tuple

import numpy as np

from bokeh.layouts import column
from bokeh.plotting import figure, Figure
from bokeh.models import WheelZoomTool, Span
from pyscivis.visualizer.plots.dimension_handler import DimensionHandler
from pyscivis.visualizer.dataclasses.state import ProfileState


class ProfileClick(TypedDict):
    x: Optional[int]
    y: Optional[int]


class ProfileFigure:
    """
    Wraps all profile figures and allows to changed them in an organized way.
    """

    def __init__(self,
                 dimension_handler: DimensionHandler,
                 min_: float,
                 max_: float
                 ) -> None:
        """
        Dynamically create profiles for every dimension.

        Args:
            dimension_handler: DimensionHandler instance.
            min_: Current data-minimum.
            max_: Current data-maximum.
        """
        self.x: Optional[int] = None
        self.y: Optional[int] = None
        self.min, self.max = self.pad_bounds(min_, max_)
        self.dimension_handler = dimension_handler
        self.profiles = dict()

        for index, dim in dimension_handler.dims.items():
            if dim.size <= 1:
                continue
            self.profiles[index] = self._create_profile_for_dims(dim.name, dim.length)

        self.layout = column(*list(self.profiles.values())[::-1], css_classes=["profile-figures"])

    def set_state(self, state: ProfileState):
        """
        Set state by updating all profiles to display the profile at the state's coordinates.

        Args:
            state: ProfileState containing click-coordinates.
        """
        self.set_coordinates(state.x, state.y)

    def get_state(self) -> ProfileState:
        """
        Get the state the current profile click-coordinate.

        Returns:
            ProfileState containing click-coordinates.
        """
        state = ProfileState(x=self.x, y=self.y)
        return state

    def set_coordinates(self,
                        x: Optional[int],
                        y: Optional[int]
                        ) -> None:
        """
        Set x- and y-click-coordinates (typically from a tap-event) and update the profiles.

        Args:
            x: X-Coordinate.
            y: Y-Coordinate.
        """
        self.x = x
        self.y = y
        self.update()

    def update_from_axis_change(self) -> None:
        """
        Null x- and y-click-coordinates and reset the profiles.
        """
        self.x = None
        self.y = None
        self.update(True)

    def update(self, reset_profiles: bool = False) -> None:
        """
        Updates all profiles using the last x- and y-click coordinates.

        Args:
            reset_profiles: Whether to only reset the profiles.
        """
        active_y_axis, active_x_axis = self.dimension_handler._order[-2:]

        for index, profile in self.profiles.items():
            self._update_profile_vals(profile, index, active_x_axis, active_y_axis, reset=reset_profiles)

    def _update_profile_vals(self,
                             profile: Figure,
                             index: int,
                             active_x_axis: int,
                             active_y_axis: int,
                             reset: bool = False
                             ) -> None:

        if reset:
            # TODO: Reorder on axes change?
            profile.title.text = profile.name
            profile.select_one({"name": "index_indicator"}).location = None
            profile.select_one({"name": "profile_indicator"}).data_source.data = dict(x=[np.nan], y=[np.nan])
            #profile.select_one({"name": "profile_points"}).data_source.data = dict(x=[np.nan], y=[np.nan])
            return

        if self.x is None or self.y is None:
            return

        click_coords = dict(x_id=active_x_axis, x_coord=self.x, y_id=active_y_axis, y_coord=self.y)
        vals: np.ndarray = self.dimension_handler.get_profile(index, click_coords)

        data_len = len(vals)
        image_len = self.dimension_handler.dims[index].length
        ratio = image_len/data_len

        if index == active_y_axis:
            own_index = self.y
        elif index == active_x_axis:
            own_index = self.x
        else:
            own_index = self.dimension_handler._values[index]
        own_index *= ratio
        profile.select_one({"name": "index_indicator"}).location = own_index
        profile.title.text = f"{profile.name} = {round(own_index, 1):g}"
        new_xs = np.arange(len(vals))*ratio
        new_ys = vals
        data = dict(x=new_xs, y=new_ys)
        profile.select_one({"name": "profile_indicator"}).data_source.data = data
        #profile.select_one({"name": "profile_points"}).data_source.data = data

    def update_y_bounds(self,
                        start: float,
                        end: float
                        ) -> None:
        """
        Update the y-bounds of all profiles. Uses a 10% padding on both sides.

        Args:
            start: Start value for the y-range.
            end: End value for the y-range.
        """
        padded_start, padded_end = self.pad_bounds(start, end)
        for profile in self.profiles.values():
            profile.y_range.update(start=padded_start, end=padded_end, bounds=(padded_start,padded_end))

    def _create_profile_for_dims(self,
                                 name: str,
                                 length: float,
                                 # unit=None
                                 ) -> Figure:
        """ Create a profile for one specific dimension (name and length) """
        profile = figure(title=name, width=300, height=200, y_axis_location="right",
                         x_range=(0, length-1), y_range=(self.min, self.max),
                         toolbar_location=None, tools=['pan'], name=name)
        profile.x_range.bounds = (0, length-1)
        profile.y_range.bounds = (self.min, self.max)
        vline = Span(location=None, dimension='height', name="index_indicator", line_color="#848484", line_width=1)
        profile.line(x=[np.nan], y=[np.nan], name="profile_indicator")
        #dots = profile.cross(x=[np.nan],y=[np.nan], name="profile_points")
        wz_tool = WheelZoomTool(maintain_focus=False, dimensions='height')
        #h_tool = HoverTool(renderers=[dots], tooltips="@y", mode='vline')
        profile.add_layout(vline)
        profile.add_tools(wz_tool)#, h_tool)
        profile.toolbar.active_scroll = wz_tool
        return profile

    @staticmethod
    def pad_bounds(start: float,
                   end: float,
                   padding_factor: float = 0.1
                   ) -> Tuple[float, float]:
        """
        Pad the supplied bounds by multiplying the bound-difference with the supplied padding factor.

        Args:
            start: Bound-start.
            end: Bound-end.
            padding_factor: Factor for padding.

        Returns:
            Tuple of padded bounds.

        Raises:
            ValueError: If start bigger than end.
        """
        if np.isnan(start) or np.isnan(end):
            return 0, 1

        if start > end:
            raise ValueError(f"Bound-start {start} cannot be bigger than Bound-end {end}")

        if start == end:
            padding = padding_factor if padding_factor != 0 else 0.1
        else:
            padding = (end - start) * padding_factor

        return start - padding, end + padding



from typing import List, Tuple, Optional

import numpy as np
from bokeh.document import Document
from bokeh.layouts import column, row
from bokeh.models import LinearColorMapper, ColorBar, Panel, Tabs, LayoutDOM, Spacer
from bokeh.models.widgets import Toggle, RadioButtonGroup

from pyscivis.visualizer.dataclasses.config import ImageConfig
from pyscivis.visualizer.dataclasses.state import State
from pyscivis.visualizer.plots.base_plot import BasePlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv
from pyscivis.visualizer.plots.components.state import StateFigure
from pyscivis.visualizer.util import create_palettes
from pyscivis.visualizer.plots.dimension_handler import DimensionHandler
from pyscivis.visualizer.plots.components import ImageFigure, HistogramFigure, DimensionControls, PaletteControls, \
    StatisticsFigure, ProfileFigure
from pyscivis.visualizer.models import OffsetLogColorMapper, OffsetLogTickFormatter
from pyscivis.visualizer.dataclasses.parser import ParsedData


class ImagePlot(BasePlot):
    """
    Creates the visualization plots for images, organizes callbacks of components and keeps everything updated.
    """

    def __init__(self,
                 data: ParsedData,
                 config: ImageConfig,
                 use_rgba: bool = False,
                 loading_indicator: Optional[LoadingDiv] = None,
                 title: str = ""
                 ) -> None:
        """
        Set up all figures and plots using the specified configuration object.

        Args:
            data: Parsed images data.
            config: Configuration settings for plots displayed by this class.
            use_rgba: If True an 'image_rgba'-glyph will be used instead of an 'image'-glyph, also disabled palettes.
            loading_indicator: None in notebook, LoadingDiv in standalone, used to print messages in the Loading-screen.
        """
        self.document: Optional[Document] = None  # Only used by the notebook for "add_next_tick_callback"
        # Flipping the np-data array along the y-axis
        data.data = data.data[..., ::-1, ::]
        self.handler = DimensionHandler(data, config.preload_bounds, loading_indicator=loading_indicator)

        if loading_indicator is not None:
            loading_indicator.set_text("Setting up widgets...")

        img_data = self._get_data()  # Initial 2D image data

        # "Fit to frame" toggle #
        self.fit_to_frame_toggle = Toggle(label="Fit bounds to frame", css_classes=["fit-to-frame-toggle"])
        self.fit_to_frame_toggle.on_click(lambda x: self.on_fit_to_frame_click())

        # Axis sliders and changers #
        self.dim_controls = DimensionControls(self.handler.dims, self.on_dim_change, self.on_axes_change)

        _min, _max = self._get_min_max()

        # Create palettes #
        palette_size = config.palette_size
        name_func_map = dict(zip(config.palette_names, config.palette_funcs))
        palette_map = create_palettes(palette_size, name_func_map)

        # ColorBars #
        color_bars = dict()
        initial_palette = palette_map[config.initial_palette][config.initial_palette_size-1]

        # Linear ColorBar #
        if np.isnan(_min) or np.isnan(_max):
            low = high = 0
        else:
            low = _min
            high = _max
        lin_cm = LinearColorMapper(palette=initial_palette, low=low, high=high)
        lin_cb = ColorBar(color_mapper=lin_cm)
        color_bars["Linear"] = lin_cb

        # Logarithmic ColorBar #
        ls = config.log_span
        log_cm = OffsetLogColorMapper(palette=initial_palette, log_span=ls, low=low, high=high)
        log_tf = OffsetLogTickFormatter(log_span=ls, low=low, high=high)

        def link_to_tf(attr, old, new):
            setattr(log_tf, attr, new)

        log_cm.on_change("low", link_to_tf)  # making sure that the custom Log models are synchronized
        log_cm.on_change("high", link_to_tf)
        log_cb = ColorBar(color_mapper=log_cm, formatter=log_tf)
        color_bars["Log"] = log_cb

        # Creating a ColorBar chooser #
        color_bar_keys: List[str] = list(color_bars.keys())
        initial_cm_index: int = color_bar_keys.index(config.initial_color_mapper)
        if not use_rgba:
            self.cbar_radios = RadioButtonGroup(labels=color_bar_keys, active=initial_cm_index, css_classes=["cbar-radios"])
            self.cbar_radios.on_click(lambda x: self.img.change_cbar(x))
        else:
            self.cbar_radios = Spacer()

        # Grabbing ordered list of cm-references for later use #
        self.color_mappers = [cb.color_mapper for cb in color_bars.values()]

        # Getting metadata of starting image #
        metadata = self.handler.get_metadata()
        dw = metadata.x.length
        dh = metadata.y.length
        x_unit = metadata.x.unit
        y_unit = metadata.y.unit

        # Create Image&State tabs #
        self.img = ImageFigure(self.on_selection_change, self.on_tap, img_data,
                               dw, dh, color_bars, initial_cm_index, use_rgba, config, title)

        self.img.plot.xaxis.axis_label = f"{metadata.x.name} ({x_unit})"
        self.img.plot.yaxis.axis_label = f"{metadata.y.name} ({y_unit})"

        image_tab = Panel(child=self.img.layout, title="Image")
        state_tab = Panel(child=StateFigure(config.format_state,
                                            self.set_state,
                                            self.get_state),
                          title="State")
        self.main_tabs = Tabs(tabs=[image_tab, state_tab])

        # Initialising remaining components #
        self.palette_sliders = PaletteControls(self.color_mappers, palette_map, palette_size, config.initial_palette,
                                               _min, _max, width=200, height=200, sizing_mode="stretch_width")
        self.histogram = HistogramFigure(img_data, _min, _max,
                                         bins=config.histogram_bins, max_calc_size=config.max_calc_size)
        self.statistics = StatisticsFigure(img_data, metadata, max_calc_size=config.max_calc_size,
                                           sizing_mode="stretch_width", margin=(25, 0, 0, 0))
        self.profile_figure = ProfileFigure(self.handler, _min, _max)

    def get_layout(self) -> LayoutDOM:
        """
        Compile all plots and controls into one LayoutDOM.

        Returns:
            LayoutDOM containing multiple layouts.
        """
        c_b = column(self.dim_controls.slider_layout, self.histogram.layout, sizing_mode="stretch_width", name="col_bottom")
        c = column(self.main_tabs, c_b)
        c_l = column(self.palette_sliders.layout, self.fit_to_frame_toggle, self.cbar_radios,
                     self.dim_controls.axis_select_layout, self.statistics.layout, name="col_left")
        c_r = column(self.profile_figure.layout, name="col_right")
        ro = row(c_l, c, c_r)
        return ro

    def set_state(self, state: State) -> None:
        """
        Set the state of the currently displayed app.

        Args:
            state: A State object.
        """
        self.dim_controls.set_state(state.dimension)
        self.img.set_state(state.image)
        self.profile_figure.set_state(state.profile)
        self.fit_to_frame_toggle.active = state.fit_to_frame
        self.palette_sliders.set_state(state.palette)
        self.cbar_radios.active = state.active_color_mapper

    def get_state(self) -> State:
        """
        Get the state of the currently displayed app.

        Returns:
            A State object.
        """
        dims = self.dim_controls.get_state()
        img = self.img.get_state()
        palette = self.palette_sliders.get_state()
        profile = self.profile_figure.get_state()
        fit_to_frame = self.fit_to_frame_toggle.active
        color_mapper = self.cbar_radios.active

        state = State(dimension=dims, image=img, profile=profile, palette=palette,
                      fit_to_frame=fit_to_frame, active_color_mapper=color_mapper)
        return state

    def on_tap(self,
               x: int,
               y: int
               ) -> None:
        """
        Update profiles-figures on tap.

        Note:
            This is a callback function called only from an Image object.

        Args:
            x: X-coordinate of the click in data-space.
            y: Y-coordinate of the click in data-space.
        """
        self.profile_figure.set_coordinates(x, y)
        self.img.set_crosshair(x, y)

    def on_fit_to_frame_click(self) -> None:
        """
        Fit or unfit current frame depending on if the button was toggled on or off.

        Note:
            This is a callback function called if fit_to_frame_toggle was clicked.
        """
        self.fit()

    def on_dim_change(self,
                      dim_index: int,
                      new: int
                      ) -> None:
        """
        Update figures if dimension values were changed.

        Note:
            This is a callback function called only from a DimensionControls object.

        Args:
            dim_index: Index of Dimension that was changed.
            new: New value for dimension.
        """
        self.handler.set_dim_value(dim_index, new)
        new_image = self._get_data()
        self.img.update_image(new_image)

        sel_vals = self.img.get_selected_values()

        if self.fit_to_frame_toggle.active:
            self.fit()

        self.histogram.update(new_image, sel_vals)
        self.statistics.update(sel_vals)
        self.profile_figure.update()

    def on_axes_change(self,
                       x_axis: int,
                       y_axis: int
                       ) -> None:
        """
        Update figures if axes were changed.

        Note:
            This is a callback function called only from a DimensionControls object.

        Args:
            x_axis: Index of the new x-axis.
            y_axis: Index of the new y-axis.
        """
        self.handler.change_axes(x_axis, y_axis)

        # get metadata of new axes
        metadata = self.handler.get_metadata()
        new_image = self._get_data()
        dw = metadata.x.length
        dh = metadata.y.length
        x_unit = metadata.x.unit
        y_unit = metadata.y.unit

        self.img.change_image(new_image, dw, dh)

        self.img.plot.xaxis.axis_label = f"{metadata.x.name} ({x_unit})"
        self.img.plot.yaxis.axis_label = f"{metadata.y.name} ({y_unit})"

        self.img.reset_selection()

        sel_vals = self.img.get_selected_values()

        if self.fit_to_frame_toggle.active:
            self.fit()

        self.histogram.update(new_image, sel_vals)
        self.statistics.update(sel_vals, metadata)

        self.profile_figure.update_from_axis_change()

    def on_selection_change(self) -> None:
        """
        Updates the histogram and statistics if the ROI changed.

        Note:
            This is a callback function called only from an Images object.
        """
        if self.dim_controls.animated_dimension is None:
            sel_vals = self.img.get_selected_values()
            self.histogram.update_selection(sel_vals)
            self.statistics.update(sel_vals)

    def fit(self) -> None:
        """
        Fit the various ranges to global or local min-max, depending on fit-to-frame-toggle.

        If fit-to-frame is active/pushed down use the local, else if it is inactive/pushed out use the global min-max.
        This method is called when the image changes, or the toggle is pressed.
        """
        # TODO: Maybe keep sliders at "fitted" interval on unfitting from frame? Possible QoL-Improvement
        min_, max_ = self._get_min_max()
        self.histogram.update_min_max(min_, max_)
        data = self.handler.get_data()
        selected = self.img.get_selected_values()
        self.histogram.update(data, selected)
        self.profile_figure.update_y_bounds(min_, max_)
        self.palette_sliders.reset_window_sliders(min_, max_)

        if np.isnan(min_) or np.isnan(max_):
            min_ = max_ = 0
        for cm in self.color_mappers:
            cm.update(low=min_, high=max_)

    def _get_data(self) -> np.ndarray:
        """ Just a shorthand. """
        return self.handler.get_data()

    def _get_min_max(self) -> Tuple[float, float]:
        """ Get global or local (min, max)-tuple depending on the fit-to-frame-toggle. """
        if self.fit_to_frame_toggle.active:
            return self.handler.get_local_min_max()
        else:
            return self.handler.get_data_min_max()




from typing import Callable, Union

from bokeh.models import RadioButtonGroup

import pyscivis.visualizer.util.multithreaded_npops as mt

LABEL_FUNC_MAP = {"Real": mt.real,
                  "Imaginary": mt.imag,
                  "Abs": mt.abso,
                  "Phase": mt.angle}


class ValueRadios:
    """
    Contains Radio-Buttons for every value-kind (Real, Imaginary, Abs, Phase).
    """

    def __init__(self,
                 on_value_change: Callable[[], None],
                 active_key: str
                 ) -> None:
        """
        Create radios and set up callbacks.

        Args:
            on_value_change: Callable to fire on radio-change.
            active_key: The initially active radio-button.
        """
        active_index = list(LABEL_FUNC_MAP.keys()).index(active_key)
        self.rg = RadioButtonGroup(labels=list(LABEL_FUNC_MAP.keys()), active=active_index)
        self.active_handler = list(LABEL_FUNC_MAP.values())[self.rg.active]
        self.rg.on_change("active", lambda attr, old, new: self.notify_parent())
        self.on_value_change = on_value_change

        self.layout = self.rg

    def set_state(self, number_kind: int) -> None:
        """
        Set state by activating the specified radio.

        Args:
            number_kind: Index of radio-button to be activated.
        """
        self.rg.active = number_kind

    def get_state(self) -> int:
        """
        Get the state of the current RadioGroup.

        Returns:
            Index of the active radio-button.
        """
        return self.rg.active

    def notify_parent(self) -> None:
        """
        Retrieve value handler for radio button and notify parent.
        """
        _handler = list(LABEL_FUNC_MAP.values())[self.rg.active]
        self.active_handler = _handler
        self.on_value_change()

    def set_active_handler(self, index: Union[int, str]) -> None:
        """
        Change the active radio-button and notify the parent.

        Args:
            index: Index or name of radio-button to be activated.
        """
        if isinstance(index, str):
            index = list(LABEL_FUNC_MAP.keys()).index(index)
        self.rg.active = index



from typing import Optional

from PIL import Image
import numpy as np

from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.extensions import FlatExtension
from pyscivis.visualizer.plots import ImagePlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv

"""
This module exemplifies the creation of flat extensions, i.e., un-nested ones.
"""


def _read_img_as_greyscale(path: str,
                           width: Optional[int] = None,
                           height: Optional[int] = None
                           ) -> np.ndarray:
    """
    Read a greyscale image into a numpy array.

    Args:
        width: Width to resize the image to.
        height: Height to resize the image to.
        path: Path of image.

    Returns:
        2-dimensional numpy-array containing [y, x] of the image (in this order).
    """
    img: Image.Image
    with Image.open(path) as img:
        if width is not None and height is not None:
            img = img.resize((width, height))
        img = img.convert('L')
        np_arr = np.asarray(img)

    return np_arr


class PILExtensionFlat(FlatExtension):
    """
    This extension allows loading of some PIL-supported images.
    """

    alias = "pil_flat"
    supported_files = (".jpg", ".jpeg", ".png", ".gif", ".bmp")
    file_description = "PIL-supported Images"
    multiple = True

    def create_plot(self,
                    loading_div: Optional[LoadingDiv] = None) -> ImagePlot:
        """
        Read a PIL-supported image from the specified path and return a plot with that image.

        Returns:
            An ImagePlot object.
        """
        if loading_div is not None:
            loading_div.set_text("Loading image..")

        all_images = []
        if len(self.path) > 1:
            width = height = 500
        else:
            width = height = None
        for path in self.path:  # we set multiple=True so we get a tuple of paths instead of just one string
            image = _read_img_as_greyscale(path, width, height)
            all_images.append(image)

        image_data = np.array(all_images)
        parsed = ParsedData(
            data=image_data,
            dim_names=["images", "y", "x"],
            dim_lengths=image_data.shape,
            dim_units=["px", "px", "px"]
        )

        return ImagePlot(parsed, self.config.image)




import ctypes
import collections
from typing import Any, Dict, List, Tuple, Callable, Iterator, Union

"""
This module includes miscellaneous helper functions. 
"""

def to_string(obj: Any) -> str:
    if isinstance(obj, (ctypes.Array, collections.Sequence)):  # might as well check sequence
        return ", ".join(obj)
    else:
        return str(obj)


def format_(val: Union[float, str],
            lower_bound: int = 3,
            upper_bound: int = 5
            ) -> str:
    if isinstance(val, str):
        return val

    if val < 10**-lower_bound or val > 10**upper_bound:
        return "{:.2e}".format(val)
    else:
        return "{:.2f}".format(val)


Palette = Tuple[str]


def create_palettes(length: int,
                    name_func_map: Dict[str, str]
                    ) -> Dict[str, List[Palette]]:
    """
    Take length of palettes and a name-funcname-map, call func from bokeh.palettes and create multiple palette-lists.

    Args:
        length: Size of the biggest palette
        name_func_map: Pairs of palette names and bokeh.palette-function names, e.g. {"TurboPalette": "turbo"}

    Returns:
        A dict mapping palette names to a list of palettes
    """
    import bokeh.palettes as bp

    def palette(plt_func: Callable[[int], Palette],
                size: int
                ) -> Iterator[Palette]:
        i = 0
        while i < size:
            yield plt_func(i + 1)
            i += 1

    palette_map = dict()
    for name, func_name in name_func_map.items():
        func: Callable[[int], Palette] = getattr(bp, func_name)
        palette_map[name] = list(palette(func, length))

    return palette_map


from typing import Union, Tuple, List

import numba as nb
import numpy as np
from math import isinf, inf, nan


@nb.njit(fastmath=True)
def _minmax_nan(array: Union[list, np.ndarray]) -> Tuple[float, float]:
    """ Dynamically calculate the nanmin and nanmax values of an array. """
    maximum = -inf
    minimum = inf
    for i in array:
        if i > maximum:
            maximum = i
        if i < minimum:
            minimum = i
    if isinf(minimum):
        minimum = nan
        maximum = nan
    return minimum, maximum


@nb.njit(parallel=True)
def _minmax_chunks_nan(array: Union[list, np.ndarray],
                       chunk_ranges: np.ndarray
                       ) -> Tuple[float, float]:
    """ Parallelize calling of "_minmax_nan"using chunks and return the final result. """
    n_chunks = len(chunk_ranges)
    max_results = [-inf] * n_chunks
    min_results = [inf] * n_chunks
    for i in nb.prange(n_chunks):
        start = chunk_ranges[i][0]
        end = chunk_ranges[i][1]
        chunk_minimum, chunk_maximum = _minmax_nan(array[start:end])
        min_results[i] = chunk_minimum
        max_results[i] = chunk_maximum
    return min(min_results), max(max_results)  # stdlib min/max ignore nans


def even_chunk_sizes(dividend: int,
                     divisor: int
                     ) -> List[int]:
    """ Calculate even chunk sizes. """
    quotient, remainder = divmod(dividend, divisor)
    cells = [quotient for _ in range(divisor)]
    for i in range(remainder):
        cells[i] += 1
    return cells


def even_chunk_ranges(dividend: int,
                      divisor: int
                      ) -> List[Tuple[int, int]]:
    """ Calculate chunk ranges in form of '[(start_0, end_0), ..., (start_n-1, end_n-1)]'. """
    sizes = even_chunk_sizes(dividend, divisor)
    ranges = []
    start = 0
    for s in sizes:
        end = start + s
        ranges.append((start, end))
        start = end
    return ranges


def fast_nanminmax(array: Union[list, np.ndarray],
                   n_chunks: int
                   ) -> Tuple[float, float]:
    """
    Custom function returning the equivalent of (np.nanmin, np.nanmax).
    To improve performance and parallelize both operations Numba's JIT and it's parallel-option is leveraged.

    The array is split into multiple chunks whose min-max calculation is then distributed over multiple threads.

    Args:
        array: 1D array to calculate min & max on.
        n_chunks: Amount of chunks that should be used to split the array.

    Returns:
        A (min, max)-tuple.
    """
    if array.size < 10000000:
        return np.nanmin(array), np.nanmax(array)
    chunk_ranges = np.array([
        [start, end]
        for start, end
        in even_chunk_ranges(len(array), n_chunks)
    ], dtype=np.int64)
    return _minmax_chunks_nan(array, chunk_ranges)



from bokeh.models.formatters import LogTickFormatter
from bokeh.core.properties import Float, Seq


class OffsetLogTickFormatter(LogTickFormatter):
    """
    Format ticks for a logarithmic colormapper while taking negative intervals into account.

    The ticks within the range of logspan are re-mapped to the interval of [low, high].

    Note:
        This class is required to display correct ticks for the OffsetLogColorMapper.

    See Also:
        OffsetLogColorMapper
    """

    #__implementation__ = "offset_log_tick_formatter.ts"

    low = Float(default=0, help="""
        REQUIRED! Should be set to the ColorMapper's low value (equal to min(data) if unspecified)
    """)
    high = Float(default=0, help="""
        REQUIRED! Should be set to the ColorMapper's high value (equal to max(data) if unspecified)
    """)
    log_span = Seq(Float, default=[1, 10], help="""
        This value is used for determining the tick labels. This has only
        to be changed if you change the default log_span of the corresponding
        OffsetLogColorMapper!
    """)



from typing import Dict, List


class MRDNode:
    """
    A node class as required by the pptree package.
    """

    def __init__(self,
                 text: str,
                 type_: str,
                 parent: "MRDNode" = None
                 ) -> None:
        """
        Create a MRDNode object with the supplied test, type and parent.

        Args:
            text: Text to display.
            type_: Type of this node.
            parent: MRDNode parent.
        """
        self.text = text
        self.type_ = type_
        self.children = []
        if parent:
            parent.children.append(self)


def nodeify_tree_structure(tree_structure: List[Dict[str, str]]) -> MRDNode:
    """
    Convert a tree-like structured list to a node-tree.

    Args:
        tree_structure: A List of Dictionaries containing name and type.

    Returns:
        The root MRDNode.

    """
    root = MRDNode("root", "root")
    _nodeify_tree_structure(root, tree_structure)
    return root


def _nodeify_tree_structure(parent_node: MRDNode,
                            tree_structure: List[Dict[str, str]]
                            ) -> None:
    """ Recursive helper function for nodeify_tree_structure. """
    for node_dict in tree_structure:
        if node_dict['parent'] == parent_node.text:
            node = MRDNode(node_dict['text'], node_dict['type'], parent_node)
            _nodeify_tree_structure(node, tree_structure)



import os
import itertools
import time
import socket
from dataclasses import dataclass
from typing import Optional, Sequence, Union, Iterator, Tuple

import numpy as np
from bokeh.io import output_notebook, show as bkshow
from bokeh.document.document import Document as BokehDocument
from bokeh.resources import INLINE
from pptree import print_tree as prettyprint_tree

from pyscivis.visualizer.notebook.resources.resources import load_resources
from pyscivis.visualizer.util.themes import theme_map
from pyscivis.visualizer.extensions import extension_manager as ext_manager, FlatExtension, NestedExtension
from pyscivis.visualizer.plots import ImagePlot, ComplexPlot
from pyscivis.visualizer.dataclasses.config import load_config
from pyscivis.visualizer.util import notebook_utils as util
from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.notebook.plot_wrapper import PlotWrapper

"""
This module allows the usage of pyscivis in Jupyter Notebooks or JupyterLab.
"""

ext_manager.check_extension_validity()

config_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), os.pardir, os.pardir, "config.toml")
config = load_config(config_path)


@dataclass
class NBState:
    """
    Just a wrapper for some state variables to avoid overusing `global`.
    """

    docker_flag = False
    custom_ports: Optional[Iterator[int]] = None
    custom_ports_length = -1
    config_ports: Optional[Iterator[int]] = None
    config_ports_length = -1
    autodiscover_ports = False
    autodiscover_delay = 2000
    theme = "default"
    notebook_url = "localhost:8888"


if config.notebook.ports:
    start: int = config.notebook.ports[0]
    end: int = config.notebook.ports[1] + 1
    NBState.config_ports = itertools.cycle(range(start, end))  # infinite iterator to stay within our port range
    NBState.config_ports_length = len(range(start, end))


def notebook_url(url: str) -> None:
    """
    Set the notebook-url so that bokeh knows where to display the plots.

    Args:
        url: URL of current notebook.
    """
    NBState.notebook_url = url


def set_theme(name: str) -> None:
    """
    Set a theme to be used for all future plots. Print valid theme names if an invalid name was supplied.

    Args:
        name: Name of theme to be set.

    """
    if name not in theme_map:
        valid_themes = ", ".join(list(theme_map.keys()))
        print(f"Invalid theme name '{name}', valid options are: {valid_themes}")
        return
    NBState.theme = name


def enable(offline: bool = False,
           docker: bool = False,
           autodiscover_ports: bool = False,
           ports: Optional[Sequence[int]] = None,
           ) -> None:
    """
    Load BokehJS into the current notebook and set up necessary configuration, e.g., in docker-env, whether to knock at ports before displaying plots.

    Args:
        offline: Setting this to True permits offline usage by loading local js-files, False will use a CDN.
        docker: Flag to enable use in dockerized Jupyter Notebook by falling back to custom/config ports.
        autodiscover_ports: If enabled ports will be knocked at to enable usage of port auto-detection tools
            before the actual plots are displayed.
        ports: Custom port range, e.g. `(10, 100)` (override config ports if not None)

    Notes:
        `docker` has to be set to `True` to activate the autodiscovery-functionality
            and the custom ports.
        The custom_ports variable of this module can be nulled if `ports`
            was specified and usage of config ports is later wanted.
            Set NBState.custom_ports manually for this.
        """
    load_resources()  # load local files into notebook, the dark-theme css for now

    options = dict()
    if offline:
        options['resources'] = INLINE

    if docker:
        NBState.docker_flag = True
        if ports:
            NBState.custom_ports = itertools.cycle(range(ports[0], ports[1]+1))
            NBState.custom_ports_length = len(range(ports[0], ports[1]+1))

    NBState.autodiscover_ports = autodiscover_ports

    output_notebook(**options)
    print_supported_extensions()


def print_supported_extensions():
    """
    Print the currently supported file-extensions.
    """
    print("Currently, the enabled extensions support usage of the following file-types:")
    file_exts = ext_manager.get_supported_files()
    file_exts = set([f[2] for f in file_exts])
    print(" ".join(file_exts))


def show(plot: Union[ImagePlot, ComplexPlot], n_retries=0, max_retries=0) -> None:
    """
    Wrap bokeh's show, allowing easier display of custom Plot-objects with automatic port assignment.

    If a port-range was specified it will try to iterate over the entire range to find an open port.

    Args:
        plot: The custom plot object to display.
        n_retries: Current amount of retries used.
        max_retries: Max amount of allowed retries.
    """
    def sp(doc: BokehDocument) -> None:  # adds the actual app into the specified bokeh-document
        """ Function handler creating a new document as required per bokeh.io.show. """
        layout = plot.get_layout()
        layout.css_classes = [NBState.theme]
        doc.theme = theme_map[NBState.theme]
        doc.add_root(layout)
        plot.document = doc
    if not NBState.docker_flag:  # no dockerized environment -> use standard bokeh port assignment
        bkshow(sp, notebook_url=NBState.notebook_url)
        return

    ports = None
    if NBState.custom_ports is not None:  # custom_ports have priority over config-defined ports
        ports = NBState.custom_ports
        max_retries = NBState.custom_ports_length
    elif NBState.config_ports is not None:
        ports = NBState.config_ports
        max_retries = NBState.config_ports_length

    if ports is None:  # use bokeh's default port
        bkshow(sp, notebook_url=NBState.notebook_url)
    else:
        port = next(ports)
        try:
            if NBState.autodiscover_ports:
                _knock_at_port(port, NBState.autodiscover_delay)
            bkshow(sp, notebook_url=NBState.notebook_url, port=port)
        except OSError:
            if n_retries == max_retries:
                raise OSError(f"Could not find open port in specified port range. Tried {n_retries} ports.")
            else:
                show(plot, n_retries+1, max_retries)


def load_file(f_path: str,
              extension_name: Optional[str] = None
              ) -> Optional[Union[PlotWrapper, "NotebookHandler"]]:
    """
    Load the specified file and return a PlotWrapper if it is a flat file or a more complex handler if it is nested.

    The specified extension name is used to get the corresponding extension which is then used to parse the specified
    file-path (which can consist of multiple

    Args:
        extension_name: Extension name to be used for handling the file path.
        f_path: Path to file.

    Returns:
        NotebookHandler or PlotWrapper.
    """
    try:
        ext_handler = ext_manager.get_extension_handler(f_path, extension_name, config)
    except ValueError:
        print("It looks like the specified path does not have a file-extension. "
              "Please specify the extension yourself by calling this function"
              " with the argument 'extension_name'.\n")
        _print_extension_options(ext_manager.get_extension_names_and_descriptions())
        return
    except IndexError:
        print("There are multiple valid extensions for the specified file(s). "
              "Please specify the extension yourself by calling this function"
              " with the argument 'extension_name'.\n")
        _print_extension_options(ext_manager.get_extension_names_and_descriptions())
        return

    if isinstance(ext_handler, NestedExtension):
        tree_struct = ext_handler.tree_struct
        root = util.nodeify_tree_structure(tree_struct)
        prettyprint_tree(root, nameattr="text")  # prints a nice layout of the file
        print("\n")
        return NotebookHandler(ext_handler)
    elif isinstance(ext_handler, FlatExtension):
        plot = ext_handler.create_plot()
        show(plot)
        return PlotWrapper(plot)
    else:
        raise TypeError("This should never happen")


load = load_file


def _print_extension_options(name_desc_list: Sequence[Tuple[str, str]]) -> None:
    print("Possible extension_name values are:\n")

    options = [f"'{alias}' to load {desc}" for alias, desc in name_desc_list]
    print("\n".join(options))


def _knock_at_port(port: int,
                   delay: int,
                   ) -> None:
    """
    Uses a (possibly non-open) port for a web-socket connection and wait the specified delay afterwards.

    Args:
        port: Port to knock at.
        delay: Time in milliseconds to wait after knocking, to give time for autodetection tools to forward ports.

    Raises:
        OSError: If port is currently in use.
    """

    nb_url = NBState.notebook_url
    if nb_url.startswith("http"):
        server_url = nb_url.rsplit(':', 1)[0]
    else:
        server_url = f"http://{nb_url.split(':')[0]}"
    with socket.socket() as s:
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            s.bind((server_url, port))
        except OSError as ose:
            raise ose    # if port is already in use we want to bubble the exception up to skip this port
        finally:
            s.close()

    time.sleep(delay/1000)


def display(data: np.ndarray,
            dim_names: Optional[Sequence[str]] = None,
            dim_units: Optional[Sequence[str]] = None,
            dim_lengths: Optional[Sequence[float]] = None,
            transpose: bool = True,
            title: str = ""
            ) -> PlotWrapper:
    """
    Display a user supplied image if data is real or acquisition if data is complex.

    Args:
        data: N-dimensional numpy-array.
        dim_names: Sequence of dimension-names.
        dim_units: Sequence of units for dimensions.
        dim_lengths: Sequence of length of dimensions (corresponds to their unit).
        transpose: Interpret data in [x, y, z] format if True, else treat input as [z, y, x].
        title: Optional title displayed above the main plot.

    Returns:
        A PlotWrapper object.

    Note:
        Fails if only one of `dim_units` and `dim_lengths` was specified.
        Also fails if amount of data dimensions is unequal to the length of the specified dim_* parameters.
    """

    input_errors = list()
    if dim_names is not None and data.ndim != len(dim_names):
        input_errors.append(f"`dim_names` has to have exactly one value for every dimension,\n\
                             length: {len(dim_names)}, dim-count: {data.ndim}")
    if (dim_units is None) ^ (dim_lengths is None):
        input_errors.append("Have to set both `dim_units` and `dim_names` or neither.")

    if dim_units is not None and data.ndim != len(dim_units):
        input_errors.append(f"`dim_units` has to have exactly one value for every dimension,\n\
              length: {len(dim_units)}, dim-count: {data.ndim}")
    if dim_lengths is not None and data.ndim != len(dim_lengths):
        input_errors.append(f"`dim_lengths` has to have exactly one value for every dimension,\n\
              length: {len(dim_lengths)}, dim-count: {data.ndim}")

    if input_errors:
        err_str = '\n'.join(input_errors)
        print(f"One more multiple invalid arguments were passed:\n {err_str}")
        return

    if transpose:
        data = np.transpose(data)
        if dim_names is not None:
            dim_names = dim_names[::-1]
        if dim_units is not None:
            dim_lengths = dim_lengths[::-1]  # dim_lengths can't be None at this point
            dim_units = dim_units[::-1]

    if np.iscomplexobj(data):  # complex ndarray -> acquisition
        data = np.array(data, dtype="complex64")
        parsed = parse_raw(data, dim_names, dim_lengths, dim_units)
        plot = ComplexPlot(parsed, config.image, title=title)
    else:
        parsed = parse_raw(data, dim_names, dim_lengths, dim_units)
        plot = ImagePlot(parsed, config.image, title=title)

    show(plot)
    return PlotWrapper(plot)


def parse_raw(data: np.ndarray,
              dim_names: Optional[Sequence[str]] = None,
              dim_lengths: Optional[Sequence[str]] = None,
              dim_units: Optional[Sequence[str]] = None,
              ) -> ParsedData:
    """
    Take a user-supplied np.ndarray and header and compile them into either images or a kspace.

    Args:
        data: An np.ndarray containing either complex or real values.
        dim_names: A sequence of strings resembling the dimension names of the data.
        dim_lengths: A sequence of strings resembling the dimension names of the data.
        dim_units: A sequence of strings resembling the dimension names of the data.

    Returns:
        A ParsedData object.

    """
    if dim_names is None:
        dim_names = list()
        for i in reversed(range(data.ndim)):
            dim_names.append(f"dim{i}")
    dim_names = tuple(dim_names)

    if dim_units is None:
        dim_units = ["pixel"] * data.ndim
    dim_units = tuple(dim_units)

    if dim_lengths is None:
        dim_lengths = data.shape
    dim_lengths = tuple(dim_lengths)

    return ParsedData(data, dim_names, dim_lengths, dim_units)


class NotebookHandler:
    """
    Manages ISMRMRD-files, creates&displays plots inside the notebook.
    """

    def __init__(self, handler: NestedExtension) -> None:
        """
        Set up the handler, create some shortcuts, and print valid combinations for container.

        Args:
            handler: ExtensionHandler used to parse the file and create plots.
        """
        self.handler = handler
        self.valid_combinations: Sequence[str] = handler.get_valid_leaf_type_combinations()
        self.print_valid()

    def display(self, combination: str) -> PlotWrapper:
        """
        Display a plot in a notebook cell.

        Args:
            combination: A "name:type" combination.

        Returns:
            A PlotWrapper object.
        """
        if combination not in self.valid_combinations:
            raise ValueError(f"{combination} is not a valid key. "
                             "Use NotebookHandler.print_valid to see valid combinations.")
        l_name = c_name = type_ = ""
        if combination.count(";") == 1:
            c_name, type_ = combination.split(";")
        elif combination.count(";") == 0:
            l_name = combination

        plot = self.handler.create_plot(leaf_name=l_name, container_name=c_name, container_type=type_)
        show(plot)

        return PlotWrapper(plot)

    def print_valid(self):
        print("Valid arguments for NotebookHandler.display are:\n")
        print("\n".join(self.valid_combinations))



import os
import sys
import subprocess

from pyscivis.arguments import parse_args


def main() -> None:  # entrypoint for >pyscivis
    args = parse_args()
    package_name = os.path.dirname(__file__)

    if args.configfile:
        config_path = os.path.join(package_name, "config.toml")
        print(config_path)
        sys.exit()

    optional_serve_args = []
    if args.allow_websocket_origin:
        for origin in args.allow_websocket_origin:
            optional_serve_args.append(f"--allow-websocket-origin={origin}")
    if args.port is not None:
        optional_serve_args.append(f"--port={args.port}")

    sp_call = ["bokeh", "serve", "--show", package_name, *optional_serve_args, "--args", *sys.argv[1:]]
    if args.server:
        sp_call.remove("--show")
    subprocess.call(sp_call)


if __name__ == '__main__':  # entrypoint for >python -m pyscivis
    main()




from typing import List, Any

import attr
import desert
import toml


@attr.s
class ImageConfig:
    """
    A dataclass containing various configuration options that mainly concern the displayed plots and image.

    This class should not be edited, instead use "config.toml" to edit config-values.
    Use ">pyscivis -c" to get the path to your "config.toml".
    Consider checking "config.py" to get an overview over possible arguments.

    See Also:
        Config
    """
    initial_color_mapper: str = attr.ib(default="Linear",
                                        validator=attr.validators.in_(["Linear", "Log"]))
    initial_value_kind: str = attr.ib(default="Real",
                                      validator=attr.validators.in_(["Real", "Imaginary", "Abs", "Phase"]))
    initial_palette: str = attr.ib(default="Greyscale")
    initial_palette_size: int = attr.ib(default=100,
                                        validator=attr.validators.in_(range(1, 257)))

    palette_names: List[str] = attr.ib(default=["Greyscale", "Turbo", "Inferno"])
    palette_funcs: List[str] = attr.ib(default=["grey", "turbo", "inferno"])
    palette_size: int = attr.ib(default=100,
                                validator=attr.validators.in_(range(1, 257)))

    preload_bounds: bool = attr.ib(default=False)

    log_span: List[int] = attr.ib(default=[1, 10])

    histogram_bins: int = attr.ib(default=20)
    max_calc_size: int = attr.ib(default=4000*4000)

    ds_downsample_images_threshold: int = attr.ib(default=1024)
    ds_threshold: int = attr.ib(default=512)
    ds_throttle: int = attr.ib(default=50)

    pref_height: int = attr.ib(default=450)
    min_height: int = attr.ib(default=300)
    max_height: int = attr.ib(default=500)
    min_width: int = attr.ib(default=200)
    max_width: int = attr.ib(default=600)
    border: int = attr.ib(default=50)

    format_state: bool = attr.ib(default=False)

    @histogram_bins.validator
    @ds_downsample_images_threshold.validator
    @ds_threshold.validator
    @ds_throttle.validator
    @pref_height.validator
    @max_height.validator
    @min_height.validator
    @max_width.validator
    @min_width.validator
    @border.validator
    def positive_int(self, attribute: attr.Attribute, value: int):
        """ Validator to assert that int-attributes are non-negative."""
        if value < 0:
            raise ValueError(f"'{attribute.name}'({value}) must be positive. Re-check the config-file.")

    def __attrs_post_init__(self):
        if self.max_height < self.min_height:
            raise ValueError(f"'max_height'({self.max_height}) cannot be smaller "
                             f"than 'min_height'({self.min_height}). Re-check the config-file.")
        if self.max_width < self.min_width:
            raise ValueError(f"'max_width'({self.max_width}) cannot be smaller "
                             f"than 'min_width'({self.min_width}). Re-check the config-file.")

        if self.initial_palette not in self.palette_names:
            raise ValueError("'initial_palette' must be in 'palette_names': "
                             f"{self.initial_palette} not in {self.palette_names}."
                             "Re-check the config-file")

        if self.initial_palette_size > self.palette_size:
            raise ValueError(f"'initial_palette_size'({self.initial_palette_size}) cannot be smaller "
                             f"than 'palette_size'({self.palette_size}). Re-check the config-file.")

        if len(self.palette_names) != len(self.palette_funcs):
            raise ValueError(f"Length of 'palette_names'({len(self.palette_names)}) must be equal to "
                             f"length of 'palette_funcs'({len(self.palette_funcs)}). Re-check the config-file.")

        bp = get_bokeh_palettes()
        for func in self.palette_funcs:
            try:
                getattr(bp, func)
            except AttributeError:
                raise ValueError(f"'{func}' from 'palette_funcs' does not exist in 'bokeh.palettes'. "
                                 f"Re-check the config-file.")


@attr.s
class NotebookConfig:
    """
    A dataclass containing various configuration options that concern Jupyter notebook-specific settings.

    Args:
        ports (List[int]): Port-range to be used for use in docker-image.

    See Also:
        Config
    """
    ports: List[int] = attr.ib(default=[50001, 50020])

    @ports.validator
    def valid_port_range(self, attribute: attr.Attribute, value: List[int]):
        """ Validator to make sure ports are not nonsensical, i.e, negative or port0 > port1. """
        if min(value) <= 0:
            raise ValueError(f"'{attribute.name}'({value}) must only contain positive integers."
                             f" Re-check the config-file.")
        if value[0] > value[1]:
            raise ValueError(
                f"'{attribute.name}'({value}) first value in port range cannot be bigger than second value."
                f" Re-check the config-file.")


@attr.s
class ParserConfig:
    """
    A dataclass containing various configuration options that concern parser-specific settings.

    Args:
        chunk_size (int): The size (in Bytes) of chunks to be read from Acquisitions.
            Bigger chunks might be somewhat faster while smaller chunks make it possible
            to offer more feedback on how much progress was made loading data.
        always_save_mmap (bool): [NOT IMPLEMENTED] Whether to always save parsed data in a custom mmap.
        collapse_parallel (bool): Whether to delete no-information data that has its origins in the
            parallel-imaging acceleration-factor.

    Warning: Always keep chunk_size over at least 50MByte to avoid severe performance drops.

    See Also:
        Config
    """
    chunk_size: int = attr.ib(default=500 * 1000 * 1000)  # 500 MB
    always_save_mmap: bool = attr.ib(default=False)
    collapse_parallel: bool = attr.ib(default=False)

    @chunk_size.validator
    def valid_chunk_size(self, attribute: attr.Attribute, value: int):
        """ Validator to assert that the chunk size is non-negative."""
        if value <= 0:
            raise ValueError(f"'{attribute.name}'({value}) must be positive. Re-check the config-file.")


@attr.s
class Config:
    """
    A dataclass containing the compiled configuration from multiple subconfigs.

    Args:
        image (ImageConfig): ImageConfig data for the image and its plots.
        notebook (NotebookConfig): NotebookConfig data for the notebook-app.
        parser (ParserConfig): ParserConfig data for the parser.
    """
    image: ImageConfig = attr.ib()
    notebook: NotebookConfig = attr.ib()
    parser: ParserConfig = attr.ib()


ConfigSchema = desert.schema(Config)


def load_config(file_path) -> Config:  # pragma: nocover
    """
    Load configuration file into a dataclass object.

    Args:
        file_path: File path of the configuration file.

    Returns:
        A Config object filled with data from the configuration file.
    """
    return ConfigSchema.load(toml.load(file_path))


def get_bokeh_palettes() -> Any:
    # This is necessary to avoid having a hard dependency when testing configs
    import bokeh.palettes as bp
    return bp




import argparse
import os


def parse_args() -> argparse.Namespace:
    """

    @return:
    """
    parser = argparse.ArgumentParser(description='pyscivis')
    parser.add_argument('-f', '--file', '--filename',
                        metavar='PATH',
                        type=str,
                        help='Path of file to be initially displayed')

    parser.add_argument('-n', '--noselector',
                        action='store_true',
                        help='Disable the file selector widget')

    parser.add_argument('-c', '--configfile',
                        action='store_true',
                        help='Print the path of the configuration file')

    parser.add_argument('-p', '--port',
                        type=int,
                        metavar='PORT',
                        help='The port the bokeh application should use. Bokeh-default: 5006')

    parser.add_argument('-s', '--server',
                        nargs='?',
                        default=None,  # None if no --server was entered
                        const=os.getcwd(),  # default value if --server was entered without argument
                        metavar="SHARED_DIR",
                        help='Run the application in server-mode. '
                             'The optional parameter sets the root directory containing accessible files. '
                             'Default for SHARED_DIR is the current working directory')

    parser.add_argument('-w', '--allow-websocket-origin',
                        action='append',
                        type=str,
                        metavar='HOST[:PORT]',
                        help="The bokeh server will only accept connections coming from the specified URL. "
                             "Can be specified multiple times to allow multiple origins. "
                             "Default: localhost:5006")

    return parser.parse_args()



import concurrent.futures as cf
from typing import Callable, Optional
import os

import numpy as np


def real(arr: np.ndarray) -> np.ndarray:
    """
    Calculate the real part of the specified numpy-array using multiple threads.

    Args:
        arr: Numpy-array with a complex dtype.

    Returns:
        Numpy-array containing the real part of the input array.

    Examples:
        >>> real(np.array([1+2j, 3+4j]))
        array([1., 3.])

    """
    return _tpe_func(arr, np.real)


def imag(arr: np.ndarray) -> np.ndarray:
    """
    Calculate the imaginary part of the specified numpy-array using multiple threads.

    Args:
        arr: Numpy-array with a complex dtype.

    Returns:
        Numpy-array containing the imaginary part of the input array.

    Examples:
        >>> imag(np.array([1+2j, 3+4j]))
        array([2., 4.])
    """
    return _tpe_func(arr, np.imag)


def abso(arr: np.ndarray) -> np.ndarray:
    """
    Calculate the absolute value of each element in the specified numpy-array using multiple threads.

    Args:
        arr: Numpy-array.

    Returns:
        Numpy-array containing the corresponding absolute values.

    Examples:
        >>> abso(np.array([1+2j, 3+4j]))
        array([2.23606798, 5.        ])

        >>> abso(np.array([-1, 3]))
        array([1, 3])

    Notes:
        For complex inputs the modulus (distance between origin and (real, imag) in the complex plane) is calculated.
    """
    return _tpe_func(arr, np.abs)


def angle(arr: np.ndarray) -> np.ndarray:
    """
    Calculate the angle of each complex number in the specified numpy-array using multiple threads.

    Args:
        arr: Numpy-array.

    Returns:
        Numpy-array containing the corresponding angles.

    Examples:
        >>> angle(np.ndarray([1, 1j, 1+1j]))
        array([0.        , 1.57079633, 0.78539816])

    Notes:

    """
    return _tpe_func(arr, np.angle)


def _tpe_func(arr: np.ndarray,
              func: Callable[[np.ndarray], np.ndarray],
              n_chunks: Optional[int] = None,
              fallback_threshold: int = 1000000  # experimenting indicates that 500k is where MT-overhead is amortised
              ) -> np.ndarray:
    """
    Apply the specified function to the specified sequence in a multithreaded fashion.

    This function was conceived to multithread `np.real`, `np.imag`, `np.angle`, and `np.phase`.
    Similar functions might work, but are not supported.

    Args:
        arr: Sequence that the function should be applied on.
        func: Callable to be applied on the sequence.
        n_chunks: Number of chunks the input-array will be cut into. These chunks are then fed to the multithreading-logic.
            By default the number of chunks if derived from the CPU-count. It is capped at the size of the input-array.
        fallback_threshold: Input-arrays with a size less than `fallback_threshold` are handled without multithreading.

    Returns:
        The resulting numpy-ndarray.

    Examples:
        >>> _tpe_func(np.array([1, -2, -3]), np.abs)
        array([1, 2, 3])
    """
    # use np inbuilt if array is too small
    if arr.size < fallback_threshold:  # experimenting indicates that this might be roughly where the MT overhead is amortised
        return func(arr)

    # if n_chunks was not specified: try to determine a good number from the cpu-count (or use 64 as a fallback)
    if n_chunks is None:
        n_cpus = os.cpu_count()
        n_chunks = n_cpus*16 if n_cpus is not None else 64

    n_chunks = min(n_chunks, arr.size)  # cannot have more chunks than array elements
    # chunk array into 'n_chunk' pieces
    arr_flat = arr.ravel()
    arr_chunked = np.array_split(arr_flat, n_chunks)

    # pre-allocate our array that will be filled with the partial results
    thread_results = np.empty(arr.size, dtype="float32")

    # multithreading work
    with cf.ThreadPoolExecutor() as executor:
        index = 0
        for result in executor.map(func, arr_chunked):
            thread_results[index:index+result.size] = result
            index += result.size

    # reshape back to original shape
    return thread_results.reshape(arr.shape)




from bokeh.core.properties import (String, List, Dict, Tuple, Readonly)
from bokeh.models import Widget


class Tree(Widget):
    """
    A widget to represent a tree-like structure with retractable nodes.

    This is a custom extension wrapping a 3rd party javascript library.
    Supply a flat list of dicts, like in the example below to initialize the tree.

    Examples:
        tree = [{"id": "ajson1", "parent": "#", "text": "Filename", "type": "file"},
            {"id": "ajson2", "parent": "ajson1", "text": "Container1", "type": "container"},
            {"id": "ajson3", "parent": "ajson2", "text": "images", "type": "images"},
            {"id": "ajson4", "parent": "ajson2", "text": "header", "type": "header"},
            {"id": "ajson5", "parent": "ajson1", "text": "Container2", "type": "container"},
            {"id": "ajson6", "parent": "ajson5", "text": "Container3", "type": "container"},
            {"id": "ajson7", "parent": "ajson6", "text": "acquisition", "type": "acquisitions"},
            {"id": "ajson8", "parent": "ajson6", "text": "header", "type": "header"}]
        filetree = Tree(tree=tree)

    """

    # __implementation__ = "filetree.ts"

    # selected leaf node's parent name and own name and header type
    selected = Readonly(Tuple(String, String, String, default=("", "", "")))
    theme = String(default="light")
    # flat list used to create tree structure
    tree = List(Dict(String, String), default=[])

"""tree = [{"id": "ajson1", "parent": "#", "text": "Filename", "type": "file"},
        {"id": "ajson2", "parent": "ajson1", "text": "Container1", "type": "container"},
        {"id": "ajson3", "parent": "ajson2", "text": "images", "type": "images"},
        {"id": "ajson4", "parent": "ajson2", "text": "header", "type": "header"},
        {"id": "ajson5", "parent": "ajson1", "text": "Container2", "type": "container"},
        {"id": "ajson6", "parent": "ajson5", "text": "Container3", "type": "container"},
        {"id": "ajson7", "parent": "ajson6", "text": "acquisition", "type": "acquisitions"},
        {"id": "ajson8", "parent": "ajson6", "text": "header", "type": "header"}]

tree_widget = Tree(tree=tree, width=600, height=600)
js_cb = CustomJS(code="console.log(cb_obj.selected)")
tree_widget.js_on_change("selected", js_cb)
tree_widget.on_change("selected", lambda attr, old, new: print(new))
curdoc().add_root(tree_widget)"""



from typing import Any, Dict, Union, Optional

import numpy as np

from bokeh.models import Div
from bokeh.layouts import column
from pyscivis.visualizer.dataclasses.handler import CurrentAxesMetaData

from ...util import format_


class StatisticsFigure:
    """
    Calculate and display statistics based on the provided data and metadata.
    """

    def __init__(self,
                 arr: np.ndarray,
                 metadata: CurrentAxesMetaData,
                 max_calc_size: int,
                 **kwargs: Any
                 ) -> None:
        """
        Set up HTML-table and its stylesheet and create initial statistics.

        Args:
            arr: N-dimensional numpy-array.
            metadata: CurrentAxesMetaData.
            max_calc_size: Threshold for calculating statistics, if the data-size
                is higher than this a Placeholder will be used.
            **kwargs: Kwargs for styling the statistics-layout directly passed on to the LayoutDOM.
        """
        self.stats = dict()  # the most recent statistics are saved here for external access

        style_div = Div(text=
                        "<style type='text/css'>"
                        ".statistics-table {font-family: Arial, Helvetica, sans-serif; border-collapse: collapse}"     
                        ".statistics-table td{border: 1px solid #ddd;padding: 4px;}"        
                        ".statistics-table tr:nth-child(even){background-color: #f2f2f2;}"        
                        ".statistics-table tr:hover {background-color: #ddd;}"
                        "</style>", visible=False)
        self._text_div = Div(css_classes=["statistics-table"], sizing_mode="stretch_width")
        self._metadata = metadata
        self._max_calc_size = max_calc_size

        self.layout = column(style_div, self._text_div, **kwargs)
        self.update(arr)

    def update(self,
               arr: np.ndarray,
               metadata: Optional[CurrentAxesMetaData] = None
               ) -> None:
        """
        Update statistics.

        Args:
            arr: 2-dimensional numpy-array.
            metadata: CurrentAxesMetaData.
        """
        if metadata:
            self._metadata = metadata
        self.stats = self._get_stats_from_arr(arr)
        self._text_div.text = self._compile_text(self.stats)

    def _get_stats_from_arr(self, arr: np.ndarray) -> Dict[str, Union[str, float]]:
        flat_arr = arr.ravel()
        pixel_count = len(flat_arr)
        if pixel_count > self._max_calc_size:
            mean = std = median = iqr = "2ManyVals"
        elif len(flat_arr[~np.isnan(flat_arr)]) == 0:
            mean = std = median = iqr = "No data"
        else:
            flat_arr = flat_arr[~np.isnan(flat_arr)]
            mean = np.mean(flat_arr)
            std = np.std(flat_arr)
            median = np.median(flat_arr)
            iqr = np.subtract(*np.percentile(flat_arr, [75, 25]))

        x_unit = self._metadata.x.unit
        y_unit = self._metadata.y.unit
        x_length = self._metadata.x.length
        y_length = self._metadata.y.length
        x_size = self._metadata.x.size
        y_size = self._metadata.y.size
        area = (arr.shape[1] * x_length/x_size) *\
               (arr.shape[0] * y_length/y_size)
        area_unit = f"{x_unit}²" if x_unit == y_unit else f"{x_unit}*{y_unit}"
        return dict({"mean": mean, "median": median, "std": std, "iqr": iqr, "pixel_count": pixel_count, "area": area, "area_unit": area_unit})

    @staticmethod
    def _compile_text(stats: Dict[str, Union[str, float]]) -> str:
        return f"""
        <table class='statistics-table'><tbody>
        <tr><td>Mean</td><td>{format_(stats["mean"])}</td></tr>
        <tr><td>Median</td><td>{format_(stats["median"])}</td></tr>
        <tr><td>Standard deviation</td><td>{format_(stats["std"])}</td></tr>
        <tr><td>IQR</td><td>{format_(stats["iqr"])}</td></tr>
        <tr><td>Amount Pixel</td><td>{stats["pixel_count"]}</td></tr>
        <tr><td>Area</td><td>{stats["area"]:.1f} {stats["area_unit"]}</td></tr>
        """



try:
    from flask import Flask, render_template
except ImportError:
    raise ImportError("flask has to be installed to run this script.")
from bokeh.embed import server_document

application = Flask(__name__)


@application.route("/")
def index():
    script = server_document(url="http://localhost:5006/pyscivis")
    return render_template("embed.html", script=script)


if __name__ == "__main__":
    application.run(port=8080)



import time

import numpy

from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.extensions import FlatExtension
from pyscivis.visualizer.plots import ImagePlot


class ExampleFlatExtension(FlatExtension):  # pragma: nocover

    alias = "example_flat"
    multiple = False
    supported_files = "*"
    file_description = "Read filepath into ASCII"

    def create_plot(self, loading_div=None):

        # This allows us to print messages to a notebook
        # or a (standalone's) loading screen with the same function
        if loading_div is None:  # invoked from a notebook
            print_msg = print
        else:  # invoked from the standalone app
            print_msg = loading_div.set_text

        # Display text while we are "calculating" the plot
        print_msg("I am currently very busy.")
        time.sleep(3)

        # Here we act as if `multiple=False` was set as a class variable
        # so self.path is a string (as opposed to a tuple of strings).
        # Now: Convert the selected file-path to a list containing its ASCII-equivalent
        x_axis_values = [ord(c) for c in self.path]

        height = 50
        amount_images = 5

        multidim_image = list()

        # Create a 3D Image
        for _ in range(amount_images):
            two_dim_image = list()
            for y in range(height):
                x_vals = numpy.roll(x_axis_values, y)  # shift array by y
                two_dim_image.append(x_vals)
            multidim_image.append(two_dim_image)

        multidim_image = numpy.array(multidim_image)

        # The Plot expects to be supplied with a ParsedData object:
        parsed = ParsedData(
            data=multidim_image,
            dim_names=["images", "y", "x"],
            dim_lengths=multidim_image.shape,
            dim_units=["image", "char", "char"]  # units for the axes
        )

        # We return an ImagePlot because our data is not complex
        return ImagePlot(parsed, self.config.image)

        # Check Section 3 for information on what kind of Plots we can display



from .image_plot import ImagePlot
from .complex_plot import ComplexPlot
from .dimension_handler import DimensionHandler
from .table_plot import TablePlot

__all__ = ["ImagePlot", "ComplexPlot", "TablePlot", "DimensionHandler"]



from typing import Tuple, Generator, List, Callable

import numpy as np
import math
from warnings import warn

from ismrmrd.xsd.ismrmrdschema.ismrmrd import ismrmrdHeader as AcquisitionsHeader
from ismrmrd.acquisition import Acquisition, AcquisitionHeader
from ismrmrd.file import Acquisitions
from pyscivis.visualizer.dataclasses.config import ParserConfig

from .base_parser import BaseParser
from pyscivis.visualizer.dataclasses.parser import ParsedData

default_acquisition_dimension_names = ("AVERAGE", "SLICE", "CONTRAST", "PHASE", "REPETITION",
                                       "SET", "SEGMENT", "COIL", "Z", "Y", "X",)

ISMRMRD_ACQ_IS_NOISE_MEASUREMENT = 19


# Utility function that generates a (start,stop) tuple for each chunk
def chunk_ranges(total: int, size: int) -> Generator[Tuple[int, int], None, None]:
    yield from zip(range(0, total, size), range(size, total+size, size))


class AcquisitionParser(BaseParser):
    """
    A parser to extract a kspace and its metadata from either ismrmrd-Aquisitions or complex numpy-arrays.
    """

    @staticmethod
    def parse_with_header(acq_data: Acquisitions,
                          header: AcquisitionsHeader,
                          config: ParserConfig,
                          printer: Callable[[str], None]
                          ) -> ParsedData:
        """
        Take Acquisition-data from ismrmrd-files and compile a kspace.

        First, the dimensions of the kspace are retrieved by reading the header.
        These dimensions are then used to create a np.ndarray filled with NaNs.
        The Acquisitions-object is then iterated over chunk-wise to limit RAM-consumption.
        Then the dimensions of each Acquisition is determined and its data is written into
        the np.ndarray at that specific dimension.
        Finally the kspace and relevant header information is returned

        Args:
            acq_data: An ismrmrd-Aquisitions object.
            header: An ismrmrd-AcquisitionHeader containing relevant metadata.
            config: An object containing the parser configuration.
            printer: A callable that accepts a string and displays it.

        Returns:
            A ParsedData object.
        """
        try:
            chunk_nbytes = config.chunk_size
            collapse_parallel = config.collapse_parallel
            shape_x = None
            encoding = header.encoding[0]
            for acq in acq_data:
                if not acq.is_flag_set(ISMRMRD_ACQ_IS_NOISE_MEASUREMENT):
                    shape_x = len(acq.data[0])
                    break

            if shape_x is None:
                raise ValueError("No non-noise measurement in acquisitions")

            # required attribute, want an exception
            shape_y = encoding.encodingLimits.kspace_encoding_step_1.maximum + 1
            try:
                shape_z = encoding.encodingLimits.kspace_encoding_step_2.maximum + 1
            except AttributeError:
                shape_z = 1

            try:
                enc_s_1_acc = encoding.parallelImaging.accelerationFactor.kspace_encoding_step_1
                enc_s_2_acc = encoding.parallelImaging.accelerationFactor.kspace_encoding_step_2
            except AttributeError:
                enc_s_1_acc = 1
                enc_s_2_acc = 1

            if collapse_parallel:
                shape_y = math.ceil(shape_y/enc_s_1_acc)
                shape_z = math.ceil(shape_z/enc_s_2_acc)

            try:
                shape_avg = encoding.encodingLimits.average.maximum + 1
            except AttributeError:
                shape_avg = 1
            try:
                shape_slice = encoding.encodingLimits.slice.maximum + 1
            except AttributeError:
                shape_slice = 1
            try:
                shape_contrast = encoding.encodingLimits.contrast.maximum + 1
            except AttributeError:
                shape_contrast = 1
            try:
                shape_phase = encoding.encodingLimits.phase.maximum + 1
            except AttributeError:
                shape_phase = 1
            try:
                shape_repetition = encoding.encodingLimits.repetition.maximum + 1
            except AttributeError:
                shape_repetition = 1
            try:
                shape_set = encoding.encodingLimits.set.maximum + 1
            except AttributeError:
                shape_set = 1

            shape_segment = 1  # turning segments into a whole
            shape_coils = acq_data[0].active_channels

            kspace = np.full((shape_avg, shape_slice, shape_contrast, shape_phase,
                              shape_repetition, shape_set, shape_segment, shape_z,
                              shape_y, shape_coils, shape_x),
                             fill_value=np.nan, dtype="complex64")

            def get_acquisition_dimensions(acq_header: AcquisitionHeader):
                """Retrieving acquisition dimensions from header."""

                enc_stp1 = acq_header.idx.kspace_encode_step_1
                enc_stp2 = acq_header.idx.kspace_encode_step_2

                # adjusting y&z's since skipped data was deleted before
                if collapse_parallel:
                    enc_stp1 //= enc_s_1_acc
                    enc_stp2 //= enc_s_2_acc

                avg = acq_header.idx.average
                slc = acq_header.idx.slice
                cntrst = acq_header.idx.contrast
                phs = acq_header.idx.phase
                rpttn = acq_header.idx.repetition
                st = acq_header.idx.set
                sgmnt = 0  # We ignore segment as we want to reassemble the kspace # TODO: only overwrite nans
                return avg, slc, cntrst, phs, rpttn, st, sgmnt, enc_stp2, enc_stp1

            nrows = len(acq_data)  # number of acq-rows
            row_nbytes = acq_data[0].data.nbytes  # size of one row in bytes
            if chunk_nbytes < row_nbytes:
                warn("chunk bytesize set lower than row bytesize; resetting")
                chunk_nbytes = row_nbytes
            chunk_nrows = chunk_nbytes // row_nbytes  # rows per chunk

            # Used for printing current state
            cur_chunk = 0
            n_chunks = math.ceil(nrows/chunk_nrows)  # number of chunks that will be loaded

            for start, stop in chunk_ranges(nrows, chunk_nrows):
                cur_chunk += 1
                printer(f"Loading Chunk {cur_chunk} of {n_chunks}...")
                chunk: List[Acquisition] = acq_data[start:stop]  # grabbing part of acq-rows
                for acq in chunk:
                    if acq.is_flag_set(ISMRMRD_ACQ_IS_NOISE_MEASUREMENT):
                        continue
                    acq_dim = get_acquisition_dimensions(acq.getHead())
                    kspace[acq_dim] = acq.data  # filling kspace with row-data

            # we transpose our current data
            # ("AVERAGE", "SLICE", "CONTRAST", "PHASE", "REPETITION", "SET", "SEGMENT", "Z", "Y", "COIL", "X")
            # to this
            # ("AVERAGE", "SLICE", "CONTRAST", "PHASE", "REPETITION", "SET", "SEGMENT", "COIL", "Z", "Y", "X")
            transpose_order = (0, 1, 2, 3, 4, 5, 6, 9, 7, 8, 10)
            kspace = np.transpose(kspace, transpose_order)
            units = ("pixel",)*11
            return ParsedData(kspace, default_acquisition_dimension_names, kspace.shape, units)
        except Exception as ex:
            printer("Something went wrong...")
            raise ex



from abc import ABC, abstractmethod
from functools import partial
from typing import Union, Sequence, Optional, Callable

from ismrmrd.xsd.ismrmrdschema.ismrmrd import ismrmrdHeader as AcquisitionsHeader
from ismrmrd import ImageHeader
from ismrmrd.file import Acquisitions, Images

from pyscivis.visualizer.dataclasses.config import ParserConfig
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv

from pyscivis.visualizer.dataclasses.parser import ParsedData

import numpy as np


class BaseParser(ABC):
    """
    The base class for Parsers that enforces "parse_with_header" to be implemented.
    """

    @classmethod
    def parse(cls,
              data: Union[np.ndarray, Images, Acquisitions],
              header: Optional[Union[Sequence[str], ImageHeader, AcquisitionsHeader]],
              config: ParserConfig,
              loading_indicator: Optional[LoadingDiv] = None
              ) -> ParsedData:
        """
        The main method to be called when there is data to be parsed.

        Depending on the supplied data-type either a series of images or a kspace is compiled into a ParsedData object.
        This method chooses the right parsing-method depending on the type of data and takes care of giving the correct
        printing-callable.

        Args:
            data: An ismrmrd-File-Container or a numpy-array.
            header: An ismrmrd-header belonging to the ismrmrd-File-Container or a Sequence of
             strings with names for the dimensions of a numpy array.
            config: An object containing the parser configuration.
            loading_indicator: A bokeh-layout-object that can display the current state graphically.

        Returns:
            A ParsedData object.
        """
        if loading_indicator is not None:
            printer = loading_indicator.set_text
        else:
            printer = partial(print, end="\r")

        return cls.parse_with_header(data, header, config, printer)

    @staticmethod
    @abstractmethod
    def parse_with_header(data: Union[Images, Acquisitions],
                          header: Union[ImageHeader, AcquisitionsHeader],
                          config: ParserConfig,
                          printer: Callable[[str], None]
                          ) -> ParsedData:
        """
        Abstract method meant to be overridden by subclasses.

        Args:
            data: An ismrmrd-File-Container.
            header: An ismrmrd-header belonging to the ismrmrd-File-Container.
            config: An object containing the parser configuration.
            printer: A callable that accepts a string and displays it.

        Raises:
            NotImplementedError: This method is meant to be overridden.
        """
        raise NotImplementedError()




import math
from typing import Optional

from bokeh.models import LayoutDOM
from pyscivis.visualizer.dataclasses.config import ImageConfig
from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.dataclasses.state import State
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv

from . import ImagePlot
from .components import ValueRadios


class ComplexPlot(ImagePlot):
    """
    Creates the visualization plots for acquisitions, organizes callbacks of components and keeps everything updated.
    """

    def __init__(self,
                 data: ParsedData,
                 config: ImageConfig,
                 loading_indicator: Optional[LoadingDiv] = None,
                 title: str = ""
                 ) -> None:
        """
        Call constructor of Superclass and initialize value-handler controls.

        Args:
            data: Parsed acquisition data.
            config: Configuration settings for plots displayed by this class.
            loading_indicator: None in notebook, LoadingDiv in standalone, used to print messages in the Loading-screen.
        """
        super().__init__(data, config, loading_indicator=loading_indicator, title=title)
        if loading_indicator is not None:
            loading_indicator.set_text("Setting up complex-value toggles...")
        vk = config.initial_value_kind
        self.value_controls = ValueRadios(self.on_value_change, vk)
        if self.handler.value_handler != self.value_controls.active_handler:
            self.on_value_change()

    def get_layout(self) -> LayoutDOM:
        """
        Compile all plots and controls into one LayoutDOM.

        Returns:
            LayoutDOM containing multiple layouts.
        """
        layout = super().get_layout()
        layout.select_one({"name": "col_left"}).children.insert(-1, self.value_controls.layout)

        return layout

    def set_state(self, state: State) -> None:
        """
        Set the state of the currently displayed app from a specified State object.

        Most state information is set by calling the superclass' `set_state`.
        Here, only the state of the value-controls is set.

        Args:
            state: State to be set.
        """
        self.value_controls.set_state(state.value_kind)
        super().set_state(state)

    def get_state(self) -> State:
        """
        Get the state of the currently displayed app.

        Most state information is retrieved by calling the superclass' `get_state`.
        Here, only the state of the value-controls are retrieved.

        Returns:
            A State object.
        """
        state: State = super().get_state()
        state.value_kind = self.value_controls.get_state()
        return state

    def on_value_change(self) -> None:
        """
        This method updates most figures when the value-handler controls were changed.
        """
        self.handler.value_handler = self.value_controls.active_handler
        new_image = self._get_data()
        min_, max_ = self._get_min_max()
        self.img.update_image(new_image)
        self.palette_sliders.reset_window_sliders(min_, max_)
        sel_vals = self.img.get_selected_values()
        self.histogram.update_min_max(min_, max_)
        self.histogram.update(new_image, sel_vals)
        self.statistics.update(sel_vals)
        self.profile_figure.update_y_bounds(min_, max_)
        self.profile_figure.update()

        if math.isnan(min_) or math.isnan(max_):
            min_ = max_ = 0
        for cm in self.color_mappers:
            cm.update(low=min_, high=max_)



import os
from typing import Dict, List, Tuple, Callable
import ismrmrd

"""
This module includes various utility functions for parsing 
ISMRMRD-Files and converting them to python-native datastructures.
"""


def extract_data_leaves(paths: Tuple[str, ...],
                        file_handler: Callable[[str], ismrmrd.File]
                        ) -> Dict[str, ismrmrd.file.Container]:
    """
    Extracts only the leaves of an h5-file and enhances them with additional information.

    Args:
        paths: Paths of ismrmrd-files.
        file_handler: Callable accepting a path-str and returning a file-handler.

    Returns:
        A list of dictionaries mapping leaf-names to H5-Containers.

    """
    leaf_dict = dict()
    for path in paths:
        ismrmrd_file = file_handler(path)
        file_basename = os.path.basename(path)
        _extract_data_leaves(leaf_dict, ismrmrd_file, "root_"+file_basename)
    return leaf_dict


def _extract_data_leaves(leaf_dict: Dict[str, ismrmrd.file.Container],
                         container: ismrmrd.file.Container,
                         parent_name: str
                         ) -> None:
    """ Recursive helper function for extract_data_leaves. """
    for child_name in container:
        child = container[child_name]
        headers = child.available()
        own_name = parent_name+"_"+child_name
        if headers:
            leaf_dict[own_name] = child
        else:
            _extract_data_leaves(leaf_dict, container[child_name], own_name)


"""
[
    {"id": "root_ajson1", "parent": "#", "text": "Filename"},
    {"id": "root_ajson1_ajson2", "parent": "root_ajson1", "text": "Container1", "type": "container"},
    {"id": "root_ajson1_ajson2_ajson3", "parent": "root_ajson1_ajson2", "text": "Image1", "type": "image"},
    {"id": "root_ajson1_ajson2_ajson4", "parent": "root_ajson1_ajson2", "text": "Header1", "type": "header"},
    {"id": "root_ajson1_ajson5", "parent": "root_ajson1", "text": "Container2", "type": "container"},
]
"""


def extract_tree_structure(paths: Tuple[str],
                           file_handler: Callable[[str], ismrmrd.File]
                           ) -> List[Dict[str, str]]:
    """
    Extract the file structure of a h5-File into a flat list.

    Args:
        paths:  Paths of ismrmrd-files.
        file_handler: Callable accepting a path-str and returning a file-handler.

    Returns:
        A list of dictionaries containing meta-information and their hierarchical parent.

    """

    ret = list()
    root = {"id": "root", "parent": "#", "text": "Files", "type": "file"}
    ret.append(root)
    for path in paths:
        ismrmrd_file = file_handler(path)
        file_basename = os.path.basename(path)
        id_ = "root_"+file_basename
        file = {"id": id_, "parent": "root", "text": file_basename, "type": "file"}
        ret.append(file)
        for child_key in ismrmrd_file:
            _extract_tree_structure(ret, ismrmrd_file[child_key], child_key, id_)
    return ret


def _extract_tree_structure(node_list: List[Dict[str, str]],
                            container: ismrmrd.file.Container,
                            container_name: str,
                            parent_name: str
                            ) -> None:
    """ Recursive helper function for extract_tree_structure. """
    id_ = parent_name+"_"+container_name
    cur = {"id": id_, "parent": parent_name, "text": container_name, "type": "container"}
    node_list.append(cur)

    headers = container.available()
    for header in headers:
        h = {"id": id_+"_"+header, "parent": id_, "text": header, "type": header}
        node_list.append(h)

    for child_key in container:
        _extract_tree_structure(node_list, container[child_key], child_key, id_)


leaf_types = ["images", "acquisitions", "header", "waveforms"]




import os
from typing import Optional, Tuple, List, Dict

from PIL import Image
import numpy as np

from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.extensions import NestedExtension
from pyscivis.visualizer.plots import ImagePlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv

"""
This module exemplifies the creation of nested extensions, using "flat" files.
"""


class PILExtensionNested(NestedExtension):
    """
    This extension allows loading of some PIL-supported images.
    """

    alias = "pil_nested"
    supported_files = (".jpg", ".jpeg", ".png", ".gif", ".bmp")
    file_description = "PIL-supported Images (Nested)"
    multiple = True

    def create_plot(self,
                    leaf_name: str,
                    container_name: str,
                    container_type: str,
                    loading_div: Optional[LoadingDiv] = None
                    ) -> ImagePlot:
        """
        Read a PIL-supported image from the specified path and return a plot with that image.

        Args:
            leaf_name: Full ID of clicked leaf, in this case the path of the image.
            container_name: Name of container with plot-data, not used here since we address images by the path.
            container_type: Type of the container, not used since all plots are images.
            loading_div: Used to notify user of progress, if None -> stdout.

        Returns:
            An ImagePlot object.
        """
        if loading_div is not None:
            loading_div.set_text("Loading image..")

        pil_image: Image.Image = self.data_leaves[leaf_name]
        pil_image = pil_image.convert('L')
        np_arr = np.asarray(pil_image)

        parsed = ParsedData(
            data=np_arr,
            dim_names=["y", "x"],
            dim_lengths=np_arr.shape,
            dim_units=["px", "px"]
        )

        return ImagePlot(parsed, self.config.image)

    def get_valid_leaf_type_combinations(self) -> List[str]:
        return [name for name in self.data_leaves.keys()]

    @staticmethod
    def get_tree_structure(file_path: Tuple[str]) -> List[Dict[str, str]]:
        """
        Create a tree structure for the specified files.

        In this case we create a tree with depth 2, the Root and its (image-)children.

        Args:
            file_path: File paths for images.

        Returns:
            List of Dicts containing data to fill a FileTree.

        """
        tree_structure = list()
        root_name = f"A whole {len(file_path)} images"
        root = {"id": "root", "parent": "#", "text": root_name, "type": "file"}
        tree_structure.append(root)
        for path in file_path:
            node_name = os.path.basename(path)
            node = {"id": path, "parent": "root", "text": node_name, "type": "images"}
            tree_structure.append(node)
        return tree_structure

    @staticmethod
    def get_data_leaves(file_path: Tuple[str]) -> Dict[str, Image.Image]:
        """
        Collect all leaves from the file paths into a dict, here: (filepath, Image)-dict.

        Args:
            file_path: Sequence of file paths.
        """
        data_leaves = dict()
        for path in file_path:
            data_leaves[path] = Image.open(path)
        return data_leaves



from typing import Dict, Tuple, Callable, Optional

import numpy as np
import numpy.typing as npt
from functools import lru_cache

from pyscivis.visualizer.plots.components.loading_div import LoadingDiv
from pyscivis.visualizer.plots.components.value_controls import LABEL_FUNC_MAP
from pyscivis.visualizer.dataclasses.parser import ParsedData
from pyscivis.visualizer.dataclasses.handler import Dim, CurrentAxesMetaData
from pyscivis.visualizer.util.jit_nanminmax import fast_nanminmax


class DimensionHandler:
    """
    Handle dimension-related operations.

    Operations performed by this class include:
    - getting the current image
    - changing dimensions
    - calculating minima and maxima
    - getting the profile at a certain point in direction of a specified dimension

    Public attributes are:

    value_handler (Callable[[npt.ArrayLike], npt.ArrayLike]):
    A callable that accepts a numpy-arraylike and returns a numpy-arraylike, e.g., np.real.

    dims (Dict[int, Dim])::
    A dictionary containing dimension-related data and mapping it to the dimension-index.
    """

    def __init__(self,
                 parsed: ParsedData,
                 preload_bounds: bool = False,
                 loading_indicator: Optional[LoadingDiv] = None,
                 ) -> None:
        """
        Do initial validation of the parsed-data and set up class attributes.

        Args:
            parsed (:obj:`ParsedData`): A ParsedData object containing dimension-related data.

        Raises:
            ValueError: If `parsed` has an illegal shape, e.g., less than 2 dimensions or more dim_units than dimensions.
        """
        p = parsed
        if p.data.size == 0:
            raise ValueError("'data' must not be empty.")
        if p.data.ndim < 2:
            raise ValueError("'data' must at least be 2-dimensional")
        if len(p.dim_names) != len(set(p.dim_names)):
            raise ValueError("'names' must contain unique axis-names.")
        if p.data.ndim != len(p.dim_names):
            raise ValueError("'data.ndim' and 'names' must be of same size.")
        if p.data.ndim != len(p.dim_lengths):
            raise ValueError("'data.ndim' and 'lengths' must be of same size.")
        if p.data.ndim != len(p.dim_units):
            raise ValueError("'data.ndim' and 'units' must be of same size.")

        zipped = zip(p.dim_names, p.dim_lengths, p.dim_units)

        self._orig_data = p.data  # used only for having a baseline to manipulate
        self._cur_data = p.data  # used for displaying (set by manipulating _orig_data)
        self.value_handler: Callable[[npt.ArrayLike], npt.ArrayLike] = LABEL_FUNC_MAP["Real"]
        self.dims: Dict[int, Dim] = {index: Dim(name, p.data.shape[index], length, unit) for index, (name, length, unit) in enumerate(zipped)}
        self._order = np.array([i for i in range(p.data.ndim)])
        self._values = np.array([0 for _ in p.dim_names])

        self.preload_bounds = preload_bounds
        self.loading_indicator = loading_indicator

        if preload_bounds:
            for key, func in LABEL_FUNC_MAP.items():
                if loading_indicator is not None:
                    loading_indicator.set_text(f"Pre-loading bounds for {key}...")
                self._get_data_min_max(func)  # save in cache

    def change_axes(self,
                    x_axis: int,
                    y_axis: int
                    ) -> None:
        """
        Take the ids of the new axes and change our data to have these axes last.

        First a transpose-order is determined by removing the specified ids from the original order
        and appending them to the back. This transpose-order is then used to re-order the original data.

        Args:
            x_axis: An integer resembling the index of the new x-axis in `dims`.
            y_axis: An integer resembling the index of the new y-axis in `dims`.
        """
        if x_axis == y_axis:
            raise ValueError(f"x_axis and y_axis cannot be the same. x_axis: {x_axis}, y_axis: {y_axis}")
        new_axes = [y_axis, x_axis]

        transpose_order = list(range(self._orig_data.ndim))
        transpose_order = [dim for dim in transpose_order if dim not in new_axes] + new_axes

        self._order = np.array(transpose_order)
        self._cur_data = np.transpose(self._orig_data, transpose_order)

    def set_dim_value(self,
                      dim_index: int,
                      value: int
                      ) -> None:
        """
        Set the specified dimension-index to the specified value.

        Args:
            dim_index: An integer resembling the index of the dimension to manipulate.
            value: An integer resembling the new value.
        """
        self._values[dim_index] = value

    def get_data(self) -> np.ndarray:
        """
        Return the current image.

        More precisely, the last two (transposed) dimensions are retrieved and the current value_handler
        is applied (i.e., np.real, np.angle, ..) to obtain the relevant portion of the data.

        Returns:
            An image as a 2-dimensional np.ndarray.
        """
        values = self.get_ordered_values()[:-2]
        data = self._cur_data[tuple(values)]
        return self.value_handler(data)

    def get_metadata(self) -> CurrentAxesMetaData:
        """
        Return the metadata of the currently active axes.

        Returns:
            A CurrentAxesMetaData object.
        """
        y, x = self._order[-2:]
        return CurrentAxesMetaData(x=self.dims[x], y=self.dims[y])

    def get_profile(self,
                    dim_id: int,
                    point: Dict[str, int],
                    ) -> np.ndarray:
        """
        Calculate the profile of a dimension at a given point.

        If the specified dimension index happens to be either the active x- or y-axis the profile can be interpreted
        as going along the axis.
        If the specified dimension is not active the profile means going `into` the image in direction of the dimension.

        Args:
            dim_id: The index of the dimension.
            point: A dictionary containing x- and y-coordinates used for the starting point of the profile.

        Returns:
            A 1-dimensional numpy-array containing the profile values of the specified dimension.
        """
        values = self._values.copy()
        # Set the point-coordinates for the active x- and y-dimension
        values[point['x_id']] = point['x_coord']  # no effect if x_id=dim_id
        values[point['y_id']] = point['y_coord']  # no effect if y_id=dim_id

        # Transpose the dimension to be profiled to the back of
        # our data and the dimension values, then retrieve the
        # profile by accessing the data with the dim-values
        transpose_order = list(range(self._orig_data.ndim))
        transpose_order = [dim for dim in transpose_order if dim != dim_id] + [dim_id]
        data = np.transpose(self._orig_data, transpose_order)
        values = values[transpose_order]
        return self.value_handler(data[tuple(values)[:-1]])  # [:-1] means we only get data for dim_id

    def get_ordered_values(self) -> np.ndarray:
        """
        Retrieve the current values in the current order.

        Current order means the order of the dimensions when this method is called, e.g., (1, 2, 0) ^= (y, x, z)
        To access the values for the base dimension-order, e.g., (0, 1, 2), use `_values` directly.

        Returns:
            A 1-dimensional numpy-array containing the ordered values.
        """
        return self._values[self._order]

    def get_local_min_max(self) -> Tuple[float, float]:
        """
        Calculate the minimum and the maximum of the current image (last 2 dimensions).

        Returns:
            A (min, max)-tuple.
        """
        min_ = np.nanmin(self.get_data())
        max_ = np.nanmax(self.get_data())
        return min_.item(), max_.item()

    def get_data_min_max(self) -> Tuple[float, float]:
        """
        Calculate and caches the minimum and maximum of the entire dataset.

        The return value is always the same for non-complex data.
        For complex data there is 4 different (min, max)-tuples possible:
        one for each value_handler, i.e., real, phase, angle, abs.

        Returns:
            A (min, max)-tuple.

        Notes:
            The min-max calculation is done by a custom, parallelized+chunked Numba-JIT script to perform min-max-calculation
                simultaneously. Check "pyscivis/visualizer/util/jit_nanminmax.py" for more information.
        """
        return self._get_data_min_max(self.value_handler)

    @lru_cache()
    def _get_data_min_max(self, value_handler: Callable[[npt.ArrayLike], npt.ArrayLike]) -> Tuple[float, float]:
        """ Helper method to allow caching (by explicitly specifying parameters for the decorator to use) """
        # only want to print the state if there is a loading-div and we are not currently in the process
        # of caching the min and max
        print_state_flag = self.loading_indicator is not None and not self.preload_bounds

        if print_state_flag:
            self.loading_indicator.set_text("Calculating MinMax-Bounds...")

        min_, max_ = fast_nanminmax(value_handler(self._orig_data).ravel(), 16)

        if print_state_flag:
            self.loading_indicator.hide()
        return min_, max_




from typing import Dict, Tuple, Optional, List

import attr
import desert
from json.decoder import JSONDecodeError


@attr.s
class AxesState:
    """
    A dataclass containing the name of the x- and y-axis.

    Args:
        x_axis (str): Name of the current x_axis.
        y_axis (str): Name of the current y_axis.

    See Also:
        DimensionState
    """
    x_axis: str = attr.ib()
    y_axis: str = attr.ib()


@attr.s
class DimensionState:
    """
    A dataclass containing the state of all dimensions, including the active axes.

    Args:
        axes (AxesState): AxesState of the current axes.
        dimensions (Dict[str, int]): Maps dimension names to their values.

    See Also:
        State
    """
    axes: AxesState = attr.ib()
    dimensions: Dict[str, int] = attr.ib()


@attr.s
class RangeState:
    """
    A dataclass containing the state of a BokehRange.

    Args:
        start (float): Start value of the range.
        end (float): End value of the range.
        bounds (Tuple[float, float]): (Start, End)-tuple defining hard zooming-/scrolling-bounds.

    See Also:
        ImageState
    """
    start: float = attr.ib()
    end: float = attr.ib()
    bounds: Tuple[float, float] = attr.ib(converter=tuple)


@attr.s
class ROIState:
    """
    A dataclass containing the state of the region-of-interest (drawn rectangle).

    Args:
        x (List[float]): A list containing one x-axis-coordinate.
        y (List[float]): A list containing one y-axis-coordinate.
        width (List[float]): A list containing one width-value.
        height (List[float]): A list containing one height-value.

    See Also:
        ImageState
    """
    x: List[float] = attr.ib()
    y: List[float] = attr.ib()
    width: List[float] = attr.ib()
    height: List[float] = attr.ib()

    @x.validator
    @y.validator
    @width.validator
    @height.validator
    def valid_length(self, attribute: attr.Attribute, value: List):
        """
        Check if the attribute contains exactly one or no element.

        Args:
            attribute: The attribute to check.
            value: The value to be checked.
        """
        if len(value) > 1:
            raise ValueError(f"'{attribute.name}'({value}) must be of length <= 1.")


@attr.s
class ImageState:
    """
    A dataclass containing the state of an Image.

    Args:
        x_range (RangeState): RangeState of the x-range.
        y_range (RangeState): RangeState of the y-range.
        roi (ROIState): ROIState of the region-of-interest.

    See Also:
        State
    """
    x_range: RangeState = attr.ib()
    y_range: RangeState = attr.ib()
    roi: ROIState = attr.ib()


@attr.s
class ProfileState:
    """
    A dataclass containing the state of all Profile-figures.

    Args:
        x (Optional[int]): x-Click-coordinate for profile calculation.
        y (Optional[int]): y-Click-coordinate for profile calculation.

    See Also:
        State
    """
    x: Optional[int] = attr.ib()
    y: Optional[int] = attr.ib()


@attr.s
class PaletteState:
    """
    A dataclass containing the state palette settings.

    Args:
        name (str): Name of the selected palette.
        length (int): Palette-size of all palettes.
        window (Tuple[float, float]): Windowing (start, end)-tuple-
        cutoff (Tuple[float, float]): Cutoff (start, end)-tuple.

    See Also:
        State
    """
    name: str = attr.ib()
    length: int = attr.ib()
    window: Tuple[float, float] = attr.ib(converter=tuple)
    cutoff: Tuple[float, float] = attr.ib(converter=tuple)


@attr.s
class State:
    """
    A dataclass containing the compiled state of all substates.

    Args:
        dimension (DimensionState): DimensionState of the dimensions.
        image (ImageState): ImageState of image.
        profile (ProfileState): ProfileState of profiles.
        palette (PaletteState): PaletteState of palette.
        fit_to_frame (bool): If fit-to-frame is enabled.
        active_color_mapper (int): Index of the active ColorMapper.
        value_kind (Optional[int]): Currently selected value_kind index.
    """
    dimension: DimensionState = attr.ib()
    image: ImageState = attr.ib()
    profile: ProfileState = attr.ib()
    palette: PaletteState = attr.ib()
    fit_to_frame: bool = attr.ib()
    active_color_mapper: int = attr.ib()
    value_kind: Optional[int] = attr.ib(default=None)


StateSchema = desert.schema(State)


def state_from_json(json_state) -> State:  # pragma: nocover
    """
    Convert a JSON-like string to a State object.

    Args:
        json_state: JSON-representation of State object.

    Returns:
        The parsed State object.
    """
    return StateSchema.loads(json_state)


def state_to_json(state: State, **kwargs) -> str:  # pragma: nocover
    """
    Convert a State object to JSON.

    Args:
        state: State object to be converted.
        **kwargs: Additional JSON-dumps keyword-arguments, like 'indent'.

    Returns:
        A JSON-like string.

    """
    return StateSchema.dumps(state, **kwargs)


def is_json_state(string) -> bool:
    """
    Check if a string is a valid JSON-representation of a State object.

    Args:
        string: String to be checked.

    Returns:
        True if the string is a valid JSON-representation of a State object, else False.
    """
    try:
        state_from_json(string)
        return True
    except JSONDecodeError:
        return False



from abc import ABC, abstractmethod

from bokeh.models import LayoutDOM


class BasePlot(ABC):
    """ Abstract class serving as the base for plots, enforcing implementation of required methods. """
    @abstractmethod
    def get_layout(self) -> LayoutDOM:
        """
        Return the UNIQUE layout of the plot, i.e., parts of the layout MUST NOT be reused at any point.

        Raises:
            NotImplementedError: This method is supposed to be overridden.
        """
        raise NotImplementedError



from typing import Optional

from bokeh.models import Span


class Crosshair:
    """ Wrapper around two bokeh-spans, allowing easy creation and manipulation of a crosshair. """

    def __init__(self):
        """
        Set up two bokeh-spans.
        """
        settings = dict(line_alpha=0.8, line_color="#d3d3d3")
        self.x_line = Span(dimension="width", location=None, **settings)
        self.y_line = Span(dimension="height", location=None, **settings)

    def set_coordinates(self,
                        x: float,
                        y: float
                        ) -> None:
        """
        Changes the X- and Y-coordinates of the Y- and X-span respectively.

        Args:
            x: New X-coordinate of the Y-span.
            y: New Y-coordinate of the X-span
        """
        self.x_line.location = y
        self.y_line.location = x

    def toggle_visible(self, visible: Optional[bool] = None) -> None:
        """
        Set the crosshair's visibility to the specified value. Toggle visibility if value not specified or None.

        Args:
            visible: Show Crosshair if True, hide if False, if None switch state.
        """
        if visible is None:
            visible = not self.x_line.visible

        self.x_line.visible = visible
        self.y_line.visible = visible



import attr


@attr.s
class Dim:
    """
    A dataclass containing the metadata of a dimension.

    Args:
        name (str): The name of the dimension, e.g., "X".
        size (int): The amount of elements in the dimension.
        length (float): The actual length of the dimension. Used with unit.
        unit (str): The unit for the dimension length, e.g., "mm".

    See Also:
        CurrentAxesMetaData
    """
    name: str = attr.ib()
    size: int = attr.ib()
    length: float = attr.ib()
    unit: str = attr.ib()


@attr.s
class CurrentAxesMetaData:
    """
    A dataclass containing the dimension-metadata of the currently displayed axes.

    Args:
        x (Dim): Dim metadata for the current x-axis.
        y (Dim): Dim metadata for the current y-axis.
    """
    x: Dim = attr.ib()
    y: Dim = attr.ib()



from typing import Any, Dict, List, Optional

from bokeh.models import ColumnDataSource, DataTable, TableColumn, LayoutDOM

from pyscivis.visualizer.plots.base_plot import BasePlot
from pyscivis.visualizer.plots.components.loading_div import LoadingDiv


class TablePlot(BasePlot):
    """ Create a Table from the supplied column-dictionary. """
    def __init__(self,
                 column_dict: Dict[str, List[Any]],
                 loading_indicator: Optional[LoadingDiv] = None,
                 **kwargs: Any) -> None:
        """
        Set up CDS and Table-kwargs.

        Args:
            column_dict: Dictionary containing (column-name, List of values)-key-value pairs.
            loading_indicator: None in notebook, LoadingDiv in standalone, used to print messages in the Loading-screen.
            **kwargs: Additional keyword-arguments supplied to the Table.

        Notes:
            For a list of possible keyword-arguments see https://docs.bokeh.org/en/latest/docs/reference/models/widgets.tables.html#bokeh.models.widgets.tables.DataTable .
        """
        if loading_indicator is not None:
            loading_indicator.set_text("Filling table rows...")
        source = ColumnDataSource(data=column_dict)

        columns = [
            TableColumn(field=field_name, title=field_name.capitalize()) for field_name in column_dict
        ]

        self.table_kwargs = dict(source=source, columns=columns, **kwargs)

    def get_layout(self) -> LayoutDOM:
        """
        Fill table keyword-args into DataTable-Constructor.

        Returns:
            A bokeh-DataTable object.
        """
        return DataTable(sizing_mode="stretch_both", sortable=False, reorderable=False, **self.table_kwargs)



import math
from typing import Any, Callable, TypedDict, Sequence

import numpy as np
from bokeh.models import BoxEditTool, ColumnDataSource, TapTool, CustomJS
from attr import asdict
from pyscivis.visualizer.dataclasses.state import ImageState, RangeState, ROIState
from pyscivis.visualizer.plots.components.crosshair import Crosshair

from .griddable_image import GriddableImage


class Point(TypedDict):
    x: Sequence[float]  # always length 1
    y: Sequence[float]


class ROI(TypedDict):
    x: Sequence[float]
    y: Sequence[float]
    width: Sequence[float]
    height: Sequence[float]


class ImageFigure(GriddableImage):
    """
    Takes care of mostly non-image related stuff on the images-plot, like managing the ROI and taps.
    """

    def __init__(self,
                 selection_callback: Callable[[], None],
                 tap_callback: Callable[[int, int], None],
                 *args: Any
                 ) -> None:
        """
        Set up the ROI, including its shadow-overlay and tool and the tap-tool.

        Args:
            selection_callback: Callable to be fired when the selection changes.
            tap_callback: Callable to be fired when a tap-event happens.
            *args: Image-related arguments passed onto the GriddableImage parent class.
        """
        super().__init__(*args)  # Create the actual image

        self.selected_src = ColumnDataSource(data=dict(selected=[]))
        self.img_shadow = self.plot.multi_polygons(color=['grey'], alpha=0.4)
        self.roi_renderer = self.plot.rect('x', 'y', 'width', 'height',
                                           source=ColumnDataSource(data=dict(x=[], y=[], width=[], height=[])),
                                           fill_alpha=0.0)

        self.roi_renderer.data_source.on_change("data", lambda attr, old, new:
                                                self.notify_selection_change(new, selection_callback))

        self._crosshair = Crosshair()
        self.plot.add_layout(self._crosshair.x_line)
        self.plot.add_layout(self._crosshair.y_line)

        self.be_tool = BoxEditTool(num_objects=1, renderers=[self.roi_renderer])
        # taptool = p.select(type=TapTool) for refactoring

        self.plot.add_tools(self.be_tool, TapTool())

        self._init_tap_tool_hack(tap_callback)
        self.layout = self.plot

    def set_state(self, state: ImageState) -> None:
        """
        Set the state of the plot and ROI.

        Args:
            state: A ImageState object.
        """
        self.plot.x_range.update(**asdict(state.x_range))
        self.plot.y_range.update(**asdict(state.y_range))
        self.roi_renderer.data_source.data = asdict(state.roi)

    def get_state(self) -> ImageState:
        """
        Get the ImageState including the current plot ranges and ROI-bounds.

        Returns:
            A ImageState object.
        """
        x_r = self.plot.x_range
        y_r = self.plot.y_range

        x_range = RangeState(start=x_r.start, end=x_r.end, bounds=x_r.bounds)
        y_range = RangeState(start=y_r.start, end=y_r.end, bounds=y_r.bounds)
        roi = ROIState(**self.roi_renderer.data_source.data)
        state = ImageState(x_range=x_range, y_range=y_range, roi=roi)
        return state

    def change_cbar(self, index: int) -> None:
        """
        Make the specified ColorBar visible, hide all others.

        Args:
            index: Index of ColorBar to be made visible.
        """
        for idx, cbar in enumerate(self.color_bars.values()):
            if idx == index:
                self.img_glyphs[idx].visible = True
                cbar.visible = True
            else:
                self.img_glyphs[idx].visible = False
                cbar.visible = False

    def set_crosshair(self,
                      ds_x: int,
                      ds_y: int
                      ) -> None:
        """
        Set crosshair coordinates.

        This function calculates screen-space coordinates for the supplied data-space coordinates and centers them
        on the specified pixel.

        Args:
            ds_x: Data-space X-coordinate.
            ds_y: Data-space Y-coordinate.
        """
        ss_x = (ds_x+0.5) * self.x_ratio
        ss_y = (ds_y+0.5) * self.y_ratio
        self._crosshair.toggle_visible(True)
        self._crosshair.set_coordinates(ss_x, ss_y)

    def notify_tap(self,
                   point: Point,
                   tap_callback: Callable[[int, int], None]
                   ) -> None:
        """
        Calculate screen-space to data-space coordinates and notify of tap.

        Args:
            point: Point containing x- and y-coordinates.
            tap_callback: Callable accepts x- and y-coordinates.
        """
        x = int(point['x'][0] / self.x_ratio)
        y = int(point['y'][0] / self.y_ratio)
        tap_callback(x, y)

    def notify_selection_change(self,
                                bounds: ROI,
                                selection_callback: Callable[[], None]
                                ) -> None:
        """
        Adjust the shadow-overlay of the ROI and notify the parent.

        Args:
            bounds: New ROI bounds.
            selection_callback: Callable fired on selection change.
        """
        self._adjust_roi_shadow_overlay(bounds)
        selection_callback()

    def reset_selection(self) -> None:
        """
        Delete the ROI.
        """
        self.roi_renderer.data_source.data = dict(x=[], y=[], width=[], height=[])
        # roi_renderer DS change will automatically trigger notify_selection_change/shadow adjustments

    def get_selected_values(self) -> np.ndarray:
        """
        Calculates the values that are inside the ROI.

        Returns:
            A 2-dimensional numpy-array containing the sliced image.
        """
        src_img = self.image_data
        if not self.roi_renderer.data_source.data['x']:
            return src_img
        else:
            # roi_renderer contains possibly stretched coordinates
            # (due to a pixel resembling more/less than 1 length-unit)
            bounds_stretched = self.roi_renderer.data_source.data

            # need to convert image-space to pixel-space
            x_0 = bounds_stretched['x'][0] / self.x_ratio
            width = bounds_stretched['width'][0] / self.x_ratio
            y_0 = bounds_stretched['y'][0]
            height = bounds_stretched['height'][0]

            x_start = math.floor(x_0 - 0.5*width)
            x_end = math.ceil(x_0 + 0.5*width)
            y_start = math.floor(y_0 - 0.5*height)
            y_end = math.ceil(y_0 + 0.5*height)

            if x_start < 0:
                x_start = 0
            if y_start < 0:
                y_start = 0
            # *_end does not matter as slicing already handles "index-out-range"-upper limits gracefully

            return src_img[y_start:y_end, x_start:x_end]

    def _adjust_roi_shadow_overlay(self, roi_bounds: ROI):
        """ Adjust the shadow-overlay of the ROI (everything outside the rectangle). """
        data = dict(self.img_shadow.data_source.data)
        # reset polygon bounds
        data['xs'] = [[[]]]
        data['ys'] = [[[]]]

        # if roi was not deleted (has a height)
        if roi_bounds['height']:
            img_poly_xs = [0, self.x_span, self.x_span, 0]
            img_poly_ys = [0, 0, self.y_span, self.y_span]

            x_start = roi_bounds['x'][0] - 0.5 * roi_bounds['width'][0]
            x_end = roi_bounds['x'][0] + 0.5 * roi_bounds['width'][0]
            y_start = roi_bounds['y'][0] - 0.5 * roi_bounds['height'][0]
            y_end = roi_bounds['y'][0] + 0.5 * roi_bounds['height'][0]
            hole_xs = [x_start, x_end, x_end, x_start]
            hole_ys = [y_start, y_start, y_end, y_end]
            data['xs'][0][0].extend((img_poly_xs, hole_xs))
            data['ys'][0][0].extend((img_poly_ys, hole_ys))
            self.img_shadow.visible = True

        else:  # roi deleted
            # necessary hack to not reset fill_color
            data['xs'][0][0].append([0])
            data['ys'][0][0].append([0])
            self.img_shadow.visible = False

        self.img_shadow.data_source.data = dict(data)

    def _init_tap_tool_hack(self, tap_callback: Callable[[int, int], None]) -> None:
        """ Make taps only be propagated if the tap-tool is active (and the taps are in bounds). """
        tap_tool_toggle = ColumnDataSource(data=dict())
        callback = CustomJS(args=dict(tt=self.plot.select(type=TapTool)[0], source=tap_tool_toggle,
                                      x_range=self.plot.x_range, y_range=self.plot.y_range),
                            code="""
            // check if taptool is active            
            if (!tt.active)
                return
            var x = cb_obj.x
            var y = cb_obj.y
            // check if click was inside image
            if (x < x_range.start || x >= x_range.end)
                return
            if (y < y_range.start || y >= y_range.end)
                return
            var data = {'x': [x], 'y': [y]}
            source.data = data
        """)
        self.plot.js_on_event("tap", callback)
        # deactivated because it interferes with the box-edit-tool double-taps
        # self.plot.on_event("doubletap", lambda x: self._crosshair.toggle_visible())
        tap_tool_toggle.on_change('data', lambda attr, old, new: self.notify_tap(new, tap_callback))



from .filetree import Tree
from .throttled_slider import ThrottledSlider
from .offset_log_color_mapper import OffsetLogColorMapper
from .offset_log_tick_formatter import OffsetLogTickFormatter

__all__ = ["Tree", "ThrottledSlider", "OffsetLogColorMapper", "OffsetLogTickFormatter"]


import json
import os
from typing import List


CSS_FILES = ["dark.css"]


def load_resources():
    from IPython.core.display import Javascript, display_javascript
    resources = get_wrapped_css()
    resources = [Javascript(f"$('head').append({json.dumps(res)})") for res in resources]
    display_javascript(*resources)


def get_wrapped_css() -> List[str]:
    """
    Read static CSS files and turn wrap them in HTML-script tags.

    Returns:
        List of wrapped JavaScript code.
    """
    wrapped_css = list()
    for css_file in CSS_FILES:
        path = os.path.join(os.path.dirname(__file__), css_file)

        with open(path, "r") as res:
            wrapped = "<style class='pyscivis' type='text/css'>%s</style>" % res.read()
            wrapped_css.append(wrapped)

    return wrapped_css



import os
import tkinter
from tkinter.filedialog import askopenfilename
from typing import List, Tuple


def select_file_tkinter(extension_name: str, valid_extensions: Tuple[Tuple[str, str, str], ...], multiple: bool):
    filetype_extensions: List[Tuple[str, str]] = list()
    for alias, desc, filetype in valid_extensions:
        if alias == extension_name:
            filetype_extensions = [(desc, filetype)]
    if not filetype_extensions:
        raise ValueError(f"No such extension {extension_name}")
    filetype_extensions.append(("All files", "*"))

    root = tkinter.Tk()
    root.attributes('-topmost', True)
    root.withdraw()

    return askopenfilename(filetypes=filetype_extensions,
                           multiple=multiple)


def select_file_from_list(file_root: str, valid_ext_names):
    tree_list = []
    for (dirpath, _, filenames) in os.walk(file_root):
        if dirpath == file_root:
            parent = "#"
        else:
            parent = os.path.dirname(dirpath)

        nodes = []

        dir_name = os.path.basename(dirpath)
        directory = dict(text=dir_name, id=dirpath, type="container", parent=parent)
        nodes.append(directory)
        for filename in filenames:
            file_ext = os.path.splitext(filename)[1]
            if file_ext not in valid_ext_names:
                continue
            file_path = os.path.join(dirpath, filename)
            file_node = dict(text=filename, id=file_path, type="header", parent=dirpath)
            nodes.append(file_node)

        tree_list.extend(nodes)

    return tree_list
